# ruff: noqa: F405
import inspect
import os
import time
from abc import ABC, abstractmethod

import numpy as np
import pinocchio

from .libcrocoddyl_pywrap import *  # noqa: F403
from .libcrocoddyl_pywrap import __raw_version__, __version__  # noqa: F401


def rotationMatrixFromTwoVectors(a, b):
    a_norm = np.linalg.norm(a)
    b_norm = np.linalg.norm(b)
    if a_norm == 0 or b_norm == 0:
        return np.eye(3)
    a_copy = a / a_norm
    b_copy = b / b_norm
    a_cross_b = np.cross(a_copy, b_copy, axis=0)
    s = np.linalg.norm(a_cross_b)
    if s == 0:
        return np.eye(3)
    c = np.dot(a_copy, b_copy)
    ab_skew = pinocchio.skew(a_cross_b)
    return np.eye(3) + ab_skew + np.dot(ab_skew, ab_skew) * (1 - c) / s**2


class DisplayAbstract(ABC):
    def __init__(self, robot, rate=-1, freq=1):
        self.robot = robot
        self._nv_root = 0.0
        self.rate = rate
        self.freq = freq
        # Visual properties
        self.frameTrajGroup = "world/robot/frame_trajectory"
        self.forceGroup = "world/robot/contact_forces"
        self.thrustGroup = "world/robot/thrust"
        self.frictionGroup = "world/robot/friction_cone"
        self.forceRadius = 0.015
        self.forceLength = 0.5
        self.forceColor = [1.0, 0.0, 1.0, 1.0]
        self.frictionConeScale = 0.2
        self.frictionConeColor = [0.0, 0.4, 0.79, 0.5]
        self.activeContacts = {}
        self.activeThrust = {}
        self.frictionMu = {}
        self.frameTrajColor = {}
        self.frameTrajLineWidth = 10
        self.x_axis = np.array([1.0, 0.0, 0.0])
        self.y_axis = np.array([0.0, 1.0, 0.0])
        self.z_axis = np.array([0.0, 0.0, 1.0])
        self.totalWeight = sum(
            m.mass for m in self.robot.model.inertias
        ) * np.linalg.norm(self.robot.model.gravity.linear)
        self._init = False

    def displayFromSolver(self, solver, factor=1.0):
        if not self._init:
            self.init(solver)
        dts = self.getDurationSequenceFromSolver(solver)
        fs = self.getForceTrajectoryFromSolver(solver)
        ps = self.getFrameTrajectoryFromSolver(solver)
        rs = self.getThrustTrajectoryFromSolver(solver)
        models = [*solver.problem.runningModels.tolist(), solver.problem.terminalModel]
        dts = [m.dt if hasattr(m, "differential") else 0.0 for m in models]
        self.display(solver.xs, dts, rs, ps, [], fs, [], dts, factor)

    def display(
        self, xs, us=[], rs=[], ps=[], pds=[], fs=[], ss=[], dts=[], factor=1.0
    ):
        if ps:
            self.displayFramePoses(ps)
        if not dts:
            dts = [0.0] * len(xs)
        S = 1 if self.rate <= 0 else max(len(xs) / self.rate, 1)
        for i, x in enumerate(xs):
            if not i % S:
                if rs:
                    if i != len(xs) - 1:
                        for r in rs[i]:
                            # Display the thrust forces
                            self.displayThrustForce(r)
                    else:
                        for key, _ in self.activeThrust.items():
                            thrustName = self.thrustGroup + "/" + key
                            self.setVisibility(thrustName, False)
                if fs:
                    self.activeContacts = {
                        k: False for k, c in self.activeContacts.items()
                    }
                    for f in fs[i]:
                        key = f["key"]
                        # Display the contact forces
                        self.displayContactForce(f)
                        # Display the friction cones
                        self.displayFrictionCone(f)
                        self.activeContacts[key] = True
                for key, c in self.activeContacts.items():
                    if c is False:
                        forceName = self.forceGroup + "/" + key
                        coneName = self.frictionGroup + "/" + key
                        self.setVisibility(forceName, False)
                        self.setVisibility(coneName, False)
                self.robot.display(x[: self.robot.nq])
                time.sleep(dts[i] * factor)

    def init(self, solver):
        frameNames, thrusters = [], []
        self.frameTrajNames = []
        models = [*solver.problem.runningModels.tolist(), solver.problem.terminalModel]
        datas = [*solver.problem.runningDatas.tolist(), solver.problem.terminalData]
        for i, data in enumerate(datas):
            model = models[i]
            if self._hasContacts(data):
                contact_model, _ = self._getContactModelAndData(model, data)
                for c in contact_model.todict().values():
                    if hasattr(c, "contact"):
                        name = self.robot.model.frames[c.contact.id].name
                    elif hasattr(c, "impact"):
                        name = self.robot.model.frames[c.impact.id].name
                    if name not in frameNames:
                        frameNames.append(name)
            if hasattr(model, "differential"):
                if isinstance(
                    model.differential.actuation, ActuationModelFloatingBaseThrusters
                ):
                    thrusters.append(model.differential.actuation.thrusters)
        for n in frameNames:
            frameId = self.robot.model.getFrameId(n)
            parentId = self.robot.model.frames[frameId].parent
            self.activeContacts[str(parentId)] = True
            self.frictionMu[str(parentId)] = 0.7
            self.frameTrajNames.append(str(frameId))
        rng = np.random.default_rng()
        for fr in self.frameTrajNames:
            self.frameTrajColor[fr] = list(
                np.hstack([rng.choice(range(256), size=3) / 256.0, 1.0])
            )
        for thrust in thrusters:
            for i, _ in enumerate(thrust):
                frameName = self.robot.model.frames[2].name
                frameId = self.robot.model.getFrameId(frameName)
                self.activeThrust[str(i)] = True
        self._addForceArrows()
        self._addFrameCurves()
        self._addThrustArrows()
        self._addFrictionCones()
        self._init = True

    @abstractmethod
    def setVisibility(self, name, status):
        """Display the frame pose."""
        raise NotImplementedError("Not implemented yet.")

    @abstractmethod
    def displayFramePoses(self, ps):
        """Display the frame pose"""
        raise NotImplementedError("Not implemented yet.")

    @abstractmethod
    def displayContactForce(self, f):
        """Display the contact force"""
        raise NotImplementedError("Not implemented yet.")

    @abstractmethod
    def displayThrustForce(self, r):
        """Display the thrust force"""
        raise NotImplementedError("Not implemented yet.")

    def getTimeSequenceFromSolver(self, solver):
        dts = self.getDurationSequenceFromSolver(solver)
        return self._getTimeSequence(dts)

    def getDurationSequenceFromSolver(self, solver):
        return [
            m.dt if hasattr(m, "differential") else 0.0
            for m in solver.problem.runningModels
        ]

    def getJointTorquesTrajectoryFromSolver(self, solver):
        us = []
        for i in range(solver.problem.T):
            data = solver.problem.runningDatas[i]
            if hasattr(data, "differential"):
                us.append(data.differential.multibody.actuation.tau[self._nv_root :])
            elif isinstance(data, ActionDataImpulseFwdDynamics):
                us.append(
                    np.zeros(solver.problem.runningModels[i].state.nv - self._nv_root)
                )
            else:
                us.append(solver.us[i][self._nv_root :])
        nu = solver.problem.runningModels[i].state.nv - self._nv_root
        us.append(np.zeros(nu + 1)[1:])  # TODO: temporal patch to fix bug in pybind11
        return us

    def getForceTrajectoryFromSolver(self, solver):
        if len(self.frameTrajNames) == 0:
            return None
        fs = []
        models = [*solver.problem.runningModels.tolist(), solver.problem.terminalModel]
        datas = [*solver.problem.runningDatas.tolist(), solver.problem.terminalData]
        for i, data in enumerate(datas):
            model = models[i]
            if self._hasContacts(data):
                contact_model, contact_data = self._getContactModelAndData(model, data)
                cost_model = self._getCostModel(model)
                fs.append(
                    self._getForceInformation(
                        model.state, contact_model, contact_data, cost_model
                    )
                )
        return fs

    def getThrustTrajectoryFromSolver(self, solver):
        fs = []
        for i, model in enumerate(solver.problem.runningModels.tolist()):
            data = solver.problem.runningDatas[i]
            if hasattr(model, "differential"):
                if isinstance(
                    model.differential.actuation, ActuationModelFloatingBaseThrusters
                ):
                    fc = []
                    ui = solver.us[i]
                    for t, thrust in enumerate(model.differential.actuation.thrusters):
                        frameName = self.robot.model.frames[2].name
                        frameId = self.robot.model.getFrameId(frameName)
                        pinocchio.updateFramePlacement(
                            model.differential.state.pinocchio,
                            data.differential.pinocchio,
                            frameId,
                        )
                        oMb = data.differential.pinocchio.oMf[frameId]
                        oMf = oMb.act(thrust.pose)
                        force = ui[t]
                        fc.append(
                            {
                                "key": str(t),
                                "oMf": oMf,
                                "f": force,
                                "type": thrust.type,
                            }
                        )
                    fs.append(fc)
        return fs

    def getFrameTrajectoryFromSolver(self, solver):
        if len(self.frameTrajNames) == 0:
            return None
        ps = {fr: [] for fr in self.frameTrajNames}
        models = [*solver.problem.runningModels.tolist(), solver.problem.terminalModel]
        datas = [*solver.problem.runningDatas.tolist(), solver.problem.terminalData]
        for key, p in ps.items():
            frameId = int(key)
            for i, data in enumerate(datas):
                model = models[i]
                if hasattr(data, "differential"):
                    # Update the frame placement if there is not contact.
                    # Note that, in non-contact cases, the action model does not compute
                    # it for efficiency reason.
                    if isinstance(data.differential, StdVec_DiffActionData):
                        differential = data.differential[0]
                    else:
                        differential = data.differential
                    if (
                        len(differential.multibody.contacts.contacts.todict().items())
                        == 0
                    ):
                        pinocchio.updateFramePlacement(
                            model.differential.state.pinocchio,
                            differential.multibody.pinocchio,
                            frameId,
                        )
                    pose = differential.multibody.pinocchio.oMf[frameId]
                    p.append(np.asarray(pose.translation.T).reshape(-1).tolist())
                elif isinstance(data, ActionDataImpulseFwdDynamics):
                    pose = data.multibody.pinocchio.oMf[frameId]
                    p.append(np.asarray(pose.translation.T).reshape(-1).tolist())
        return ps

    def _getTimeSequence(self, dts):
        t, ts = 0.0, [0.0]
        for dt in dts:
            t += dt
            ts.append(t)
        return ts

    def _hasContacts(self, data):
        if hasattr(data, "differential"):
            if isinstance(
                data.differential,
                DifferentialActionDataContactFwdDynamics,
            ) or isinstance(
                data.differential,
                DifferentialActionDataContactInvDynamics,
            ):
                return True
        elif isinstance(data, ActionDataImpulseFwdDynamics):
            return True

    def _getContactModelAndData(self, model, data):
        if hasattr(data, "differential"):
            if isinstance(
                data.differential,
                DifferentialActionDataContactFwdDynamics,
            ) or isinstance(
                data.differential,
                DifferentialActionDataContactInvDynamics,
            ):
                return (
                    model.differential.contacts.contacts,
                    data.differential.multibody.contacts.contacts,
                )
            elif isinstance(data.differential, StdVec_DiffActionData) and (
                isinstance(
                    data.differential,
                    DifferentialActionDataContactFwdDynamics,
                )
                or isinstance(
                    data.differential,
                    DifferentialActionDataContactInvDynamics,
                )
            ):
                return (
                    model.differential[0].contacts.contacts,
                    data.differential[0].multibody.contacts.contacts,
                )
        elif isinstance(data, ActionDataImpulseFwdDynamics):
            return model.impulses.impulses, data.multibody.impulses.impulses

    def _getCostModel(self, model):
        if hasattr(model, "differential"):
            return model.differential.costs.costs
        elif isinstance(model, ActionModelImpulseFwdDynamics):
            return model.costs.costs

    def _getForceInformation(self, state, contact_model, contact_data, cost_model):
        fc = []
        for key, contact in contact_data.todict().items():
            if contact_model[key].active:
                joint = state.pinocchio.frames[contact.frame].parent
                oMf = contact.pinocchio.oMi[joint] * contact.jMf
                fiMo = pinocchio.SE3(
                    contact.pinocchio.oMi[joint].rotation.T,
                    contact.jMf.translation,
                )
                force = fiMo.actInv(contact.fext)
                R = np.eye(3)
                mu = 0.7
                for c in cost_model.todict().values():
                    if isinstance(
                        c.cost.residual,
                        ResidualModelContactFrictionCone,
                    ):
                        if contact.frame == c.cost.residual.id:
                            R = c.cost.residual.reference.R
                            mu = c.cost.residual.reference.mu
                            continue
                fc.append(
                    {
                        "key": str(joint),
                        "oMf": oMf,
                        "f": force,
                        "R": R,
                        "mu": mu,
                    }
                )
        return fc


class GepettoDisplay(DisplayAbstract):
    def __init__(
        self,
        robot,
        rate=-1,
        freq=1,
        cameraTF=None,
        floor=True,
        frameNames=None,
        visibility=False,
    ):
        DisplayAbstract.__init__(self, robot, rate, freq)
        if frameNames is not None:
            print("Deprecated. Do not pass frameNames")

        # Visuals properties
        self.fullVisibility = visibility
        self.floorGroup = "world/floor"
        self.backgroundColor = [1.0, 1.0, 1.0, 1.0]
        self.floorScale = [0.5, 0.5, 0.5]
        self.floorColor = [0.7, 0.7, 0.7, 1.0]
        self.frictionConeRays = True
        self._addRobot()
        self._setBackground()
        if cameraTF is not None:
            self.robot.viewer.gui.setCameraTransform(self.robot.viz.windowID, cameraTF)
        if floor:
            self._addFloor()
        self.robot.viewer.gui.createGroup(self.forceGroup)
        self.robot.viewer.gui.createGroup(self.frictionGroup)
        self.robot.viewer.gui.createGroup(self.frameTrajGroup)
        self._addForceArrows()
        self._addThrustArrows()
        self._addFrameCurves()
        self._addFrictionCones()

    def setVisibility(self, name, status):
        self.robot.viewer[name].set_property("visible", status)
        if status:
            self.robot.viewer.gui.setVisibility(name, "ON")
        else:
            self.robot.viewer.gui.setVisibility(name, "OFF")

    def displayFramePoses(self, ps):
        for key, p in ps.items():
            self.robot.viewer.gui.setCurvePoints(self.frameTrajGroup + "/" + key, p)

    def displayContactForce(self, f):
        key, pose, wrench = f["key"], f["oMf"], f["f"]
        forceMagnitud = np.linalg.norm(wrench.linear) / self.totalWeight
        R = rotationMatrixFromTwoVectors(self.x_axis, wrench.linear)
        forcePose = pinocchio.SE3ToXYZQUATtuple(pinocchio.SE3(R, pose.translation))
        forceName = self.forceGroup + "/" + key
        self.robot.viewer.gui.applyConfiguration(forceName, forcePose)
        self.robot.viewer.gui.setVector3Property(
            forceName, "Scale", [1.0 * forceMagnitud, 1.0, 1.0]
        )
        self.robot.viewer.gui.setVisibility(forceName, "ON")

    def displayThrustForce(self, f):
        key, pose, thrust = f["key"], f["oMf"], f["f"]
        wrench = pose.act(pinocchio.Force(np.array([0, 0, thrust]), np.zeros(3)))
        forceMagnitud = np.linalg.norm(wrench.linear) / self.totalWeight
        forcePose = pinocchio.SE3ToXYZQUATtuple(pinocchio.SE3(R, pose.translation))
        thrustName = self.thrustGroup + "/" + key
        self.robot.viewer.gui.applyConfiguration(thrustName, forcePose)
        self.robot.viewer.gui.setVector3Property(
            thrustName, "Scale", [1.0 * forceMagnitud, 1.0, 1.0]
        )

    def displayFrictionCone(self, f):
        key, pose, mu = f["key"], f["oMf"], f["mu"]
        position = pose
        position.rotation = f["R"]
        frictionName = self.frictionGroup + "/" + key
        self._setConeMu(key, mu)
        self.robot.viewer.gui.applyConfiguration(
            frictionName,
            list(np.array(pinocchio.SE3ToXYZQUAT(position)).squeeze()),
        )
        self.robot.viewer.gui.setVisibility(frictionName, "ON")

    def _addRobot(self):
        # Spawn robot model
        self.robot.initViewer(windowName="crocoddyl", loadModel=False)
        self.robot.loadViewerModel(rootNodeName="robot")

    def _setBackground(self):
        # Set white background and floor
        window_id = self.robot.viewer.gui.getWindowID("crocoddyl")
        self.robot.viewer.gui.setBackgroundColor1(window_id, self.backgroundColor)
        self.robot.viewer.gui.setBackgroundColor2(window_id, self.backgroundColor)

    def _addFloor(self):
        self.robot.viewer.gui.createGroup(self.floorGroup)
        self.robot.viewer.gui.addFloor(self.floorGroup + "/flat")
        self.robot.viewer.gui.setScale(self.floorGroup + "/flat", self.floorScale)
        self.robot.viewer.gui.setColor(self.floorGroup + "/flat", self.floorColor)
        self.robot.viewer.gui.setLightingMode(self.floorGroup + "/flat", "OFF")

    def _addForceArrows(self):
        for key in self.activeContacts:
            forceName = self.forceGroup + "/" + key
            self.robot.viewer.gui.addArrow(
                forceName, self.forceRadius, self.forceLength, self.forceColor
            )
            self.robot.viewer.gui.setFloatProperty(forceName, "Alpha", 1.0)
        if self.fullVisibility:
            self.robot.viewer.gui.setVisibility(self.forceGroup, "ALWAYS_ON_TOP")

    def _addFrictionCones(self):
        for key in self.activeContacts:
            self._createCone(key, self.frictionConeScale, mu=0.7)

    def _addFrameCurves(self):
        for key in self.frameTrajNames:
            frameName = self.frameTrajGroup + "/" + key
            self.robot.viewer.gui.addCurve(
                frameName,
                [np.array([0.0, 0.0, 0.0]).tolist()] * 2,
                self.frameTrajColor[key],
            )
            self.robot.viewer.gui.setCurveLineWidth(frameName, self.frameTrajLineWidth)
            if self.fullVisibility:
                self.robot.viewer.gui.setVisibility(frameName, "ALWAYS_ON_TOP")

    def _createCone(self, coneName, scale=1.0, mu=0.7):
        m_generatrices = np.matrix(np.empty([3, 4]))
        m_generatrices[:, 0] = np.matrix([mu, mu, 1.0]).T
        m_generatrices[:, 0] = m_generatrices[:, 0] / np.linalg.norm(
            m_generatrices[:, 0]
        )
        m_generatrices[:, 1] = m_generatrices[:, 0]
        m_generatrices[0, 1] *= -1.0
        m_generatrices[:, 2] = m_generatrices[:, 0]
        m_generatrices[:2, 2] *= -1.0
        m_generatrices[:, 3] = m_generatrices[:, 0]
        m_generatrices[1, 3] *= -1.0

        v = [[0.0, 0.0, 0.0]]
        for k in range(m_generatrices.shape[1]):
            v.append(m_generatrices[:3, k].T.tolist()[0])
        v.append(m_generatrices[:3, 0].T.tolist()[0])
        coneGroup = self.frictionGroup + "/" + coneName
        self.robot.viewer.gui.createGroup(coneGroup)

        meshGroup = coneGroup + "/cone"
        self.robot.viewer.gui.addCurve(meshGroup, v, self.frictionConeColor)
        self.robot.viewer.gui.setCurveMode(meshGroup, "TRIANGLE_FAN")
        if self.frictionConeRays:
            lineGroup = coneGroup + "/lines"
            self.robot.viewer.gui.createGroup(lineGroup)
            for k in range(m_generatrices.shape[1]):
                self.robot.viewer.gui.addLine(
                    lineGroup + "/" + str(k),
                    [0.0, 0.0, 0.0],
                    m_generatrices[:3, k].T.tolist()[0],
                    self.frictionConeColor,
                )
        self.robot.viewer.gui.setScale(coneGroup, [scale, scale, scale])
        if self.fullVisibility:
            self.robot.viewer.gui.setVisibility(meshGroup, "ALWAYS_ON_TOP")
            self.robot.viewer.gui.setVisibility(lineGroup, "ALWAYS_ON_TOP")

    def _setConeMu(self, coneName, mu):
        current_mu = self.frictionMu[coneName]
        if mu != current_mu:
            self.frictionMu[coneName] = mu
            coneGroup = self.frictionGroup + "/" + coneName

            self.robot.viewer.gui.deleteNode(coneGroup, True)
            self._createCone(coneName, self.frictionConeScale, mu)


class MeshcatDisplay(DisplayAbstract):
    def __init__(
        self,
        robot,
        rate=-1,
        freq=1,
        cameraTF=None,
        floor=True,
        frameNames=None,
        visibility=True,
    ):
        DisplayAbstract.__init__(self, robot, rate, freq)
        if frameNames is not None:
            print("Deprecated. Do not pass frameNames")
        robot.setVisualizer(
            pinocchio.visualize.MeshcatVisualizer(
                model=self.robot.model,
                collision_model=self.robot.collision_model,
                visual_model=self.robot.visual_model,
            )
        )
        if cameraTF is not None and hasattr(self.robot.viz, "viewer"):
            self.robot.viewer["/Cameras/default"].set_transform(
                pinocchio.XYZQUATToSE3(cameraTF).homogeneous
            )
        self._addRobot(visibility)
        self._addForceArrows()
        self._addThrustArrows()
        self._addFrictionCones()

    def setVisibility(self, name, status):
        self.robot.viewer[name].set_property("visible", status)

    def displayFramePoses(self, ps):
        for key in ps.keys():
            vertices = np.array(ps[key]).T
            self._addFrameCurves(key, vertices)

    def displayContactForce(self, f):
        key, pose, wrench = f["key"], f["oMf"], f["f"]
        forceMagnitud = np.linalg.norm(wrench.linear) / self.totalWeight
        R = rotationMatrixFromTwoVectors(self.y_axis, wrench.linear)
        forcePose = pinocchio.SE3(
            R,
            pose.translation
            + np.dot(
                R,
                np.array([0.0, forceMagnitud * self.forceLength / 2, 0.0]),
            ),
        )
        forceName = self.forceGroup + "/" + key
        self.robot.viewer[forceName].set_property("visible", False)
        self.robot.viewer[forceName].set_transform(forcePose.homogeneous)
        self.robot.viewer[forceName].set_property("scale", [1.0, forceMagnitud, 1.0])
        self.robot.viewer[forceName].set_property("visible", True)

    def displayThrustForce(self, f):
        key, pose, thrust = f["key"], f["oMf"], f["f"]
        wrench = pose.act(pinocchio.Force(np.array([0, 0, thrust]), np.zeros(3)))
        forceMagnitud = np.linalg.norm(wrench.linear) / self.totalWeight
        R = rotationMatrixFromTwoVectors(self.y_axis, wrench.linear)
        forcePose = pinocchio.SE3(
            R,
            pose.translation
            + np.dot(
                R,
                np.array([0.0, forceMagnitud * self.forceLength / 2, 0.0]),
            ),
        )
        thrustName = self.thrustGroup + "/" + key
        self.robot.viewer[thrustName].set_property("visible", False)
        self.robot.viewer[thrustName].set_transform(forcePose.homogeneous)
        self.robot.viewer[thrustName].set_property("scale", [1.0, forceMagnitud, 1.0])
        self.robot.viewer[thrustName].set_property("visible", True)

    def displayFrictionCone(self, f):
        key, pose, mu = f["key"], f["oMf"], f["mu"]
        R = pinocchio.utils.rpyToMatrix(-np.pi / 2, 0.0, 0.0)
        conePose = pinocchio.SE3(
            R,
            pose.translation
            + np.dot(R, np.array([0.0, -self.frictionConeScale / 2, 0.0])),
        ).homogeneous
        coneName = self.frictionGroup + "/" + key
        self.robot.viewer[coneName].set_property("radiusBottom", mu)
        self.robot.viewer[coneName].set_transform(conePose)
        self.robot.viewer[coneName].set_property("visible", True)

    def _addRobot(self, openWindow):
        self.robot.initViewer(open=openWindow)
        self.robot.loadViewerModel(rootNodeName="robot")

    def _addForceArrows(self):
        import meshcat.geometry as g

        meshColor = g.MeshLambertMaterial(
            color=self._rgbToHexColor(self.forceColor[:3]), reflectivity=0.8
        )
        for key in self.activeContacts:
            forceName = self.forceGroup + "/" + key
            self.robot.viewer[forceName].set_object(
                g.Cylinder(self.forceLength, self.forceRadius), meshColor
            )

    def _addThrustArrows(self):
        import meshcat.geometry as g

        meshColor = g.MeshLambertMaterial(
            color=self._rgbToHexColor(self.forceColor[:3]), reflectivity=0.8
        )
        for key in self.activeThrust:
            thrustName = self.thrustGroup + "/" + key
            self.robot.viewer[thrustName].set_object(
                g.Cylinder(self.forceLength, self.forceRadius), meshColor
            )

    def _addFrictionCones(self):
        import meshcat.geometry as g

        meshColor = g.MeshLambertMaterial(
            color=self._rgbToHexColor(self.frictionConeColor[:3]),
            reflectivity=0.8,
            opacity=0.2,
            transparent=True,
        )
        for key in self.activeContacts:
            coneName = self.frictionGroup + "/" + key
            mu = self.frictionMu[key]
            self.robot.viewer[coneName].set_object(
                g.Cylinder(
                    self.frictionConeScale, None, 0.0, mu * self.frictionConeScale
                ),
                meshColor,
            )

    def _addFrameCurves(self, key=None, vertices=None):
        if key is None and vertices is None:
            return
        import meshcat.geometry as g

        frameName = self.frameTrajGroup + "/" + key
        meshColor = g.LineBasicMaterial(
            color=self._rgbToHexColor(self.frameTrajColor[key][:3]), linewidth=3.0
        )
        self.robot.viewer[frameName].set_object(
            g.Line(g.PointsGeometry(vertices), meshColor)
        )

    def _rgbToHexColor(self, rgbColor):
        return "0x{:02x}{:02x}{:02x}".format(
            *tuple(np.rint(255 * np.array(rgbColor)).astype(int))
        )


class RvizDisplay(DisplayAbstract):
    def __init__(
        self,
        robot,
        rate=-1,
        freq=1,
    ):
        DisplayAbstract.__init__(self, robot, rate, freq)
        # Disable garbage collection for better timings.
        import gc

        gc.disable()
        # Import ROS modules
        self.ROS_VERSION = int(os.environ["ROS_VERSION"])
        if self.ROS_VERSION == 2:
            import rclpy
        else:
            import roslaunch
            import rospy
        import crocoddyl_ros
        from urdf_parser_py.urdf import URDF

        # Init the ROS node and publishers
        if self.ROS_VERSION == 2:
            if not rclpy.ok():
                rclpy.init()
        else:
            self.roscore = roslaunch.parent.ROSLaunchParent(
                "crocoddyl_display", [], is_core=True
            )
            self.roscore.start()
            rospy.init_node("crocoddyl_display", anonymous=True)
            rospy.set_param("use_sim_time", True)
            rospy.set_param(
                "robot_description", URDF.from_xml_file(robot.urdf).to_xml_string()
            )  # TODO: hard code robot.urdf because we cannot convert a Pinocchio model into an URDF
            filename = os.path.join(
                os.path.dirname(
                    os.path.abspath(inspect.getfile(inspect.currentframe()))
                ),
                "crocoddyl.rviz",
            )
            rviz_args = [
                os.path.join(
                    os.path.dirname(
                        os.path.abspath(inspect.getfile(inspect.currentframe()))
                    ),
                    "crocoddyl.launch",
                ),
                "filename:='{filename}'".format_map(locals()),
            ]
            roslaunch_args = rviz_args[1:]
            roslaunch_file = [
                (
                    roslaunch.rlutil.resolve_launch_arguments(rviz_args)[0],
                    roslaunch_args,
                )
            ]
            self.rviz = roslaunch.parent.ROSLaunchParent(
                "crocoddyl_display", roslaunch_file, is_core=False
            )
            self.rviz.start()
        self._wsPublisher = crocoddyl_ros.WholeBodyStateRosPublisher(
            robot.model, "whole_body_state", "map"
        )
        self._wtPublisher = crocoddyl_ros.WholeBodyTrajectoryRosPublisher(
            robot.model, "whole_body_trajectory", "map"
        )
        root_id = crocoddyl_ros.getRootJointId(self.robot.model)
        self._nv_root = (
            self.robot.model.joints[root_id].nv
            if self.robot.model.frames[root_id].name != "universe"
            else 0
        )

    def __del__(self):
        if self.ROS_VERSION == 1:
            self.roscore.shutdown()
            self.rviz.shutdown()

    def displayFromSolver(self, solver, factor=1.0):
        if not self._init:
            self.init(solver)
        xs = solver.xs
        us = self.getJointTorquesTrajectoryFromSolver(solver)
        dts = self.getDurationSequenceFromSolver(solver)
        models = [*solver.problem.runningModels.tolist(), solver.problem.terminalModel]
        dts = [m.dt if hasattr(m, "differential") else 0.0 for m in models]
        ps, pds = self.getFrameTrajectoryFromSolver(solver)
        fs, ss = self.getForceTrajectoryFromSolver(solver)
        self.display(xs, us, dts, ps, pds, fs, ss, dts, factor)

    def display(
        self, xs, us=[], rs=[], ps=[], pds=[], fs=[], ss=[], dts=[], factor=1.0
    ):
        nq = self.robot.model.nq
        if not dts:
            dts = [0.01] * len(xs)
        ts = self._getTimeSequence(dts)[:-1]
        if len(ps) != 0:
            self._wtPublisher.publish(ts, xs, us, ps, pds, fs, ss)
        else:
            self._wtPublisher.publish(ts, xs, us)
        for i in range(len(xs)):
            t, x, u = ts[i], xs[i], us[i]
            if len(ps) != 0:
                p, pd = ps[i], pds[i]
                f, s = fs[i], ss[i]
                self._wsPublisher.publish(t, x[:nq], x[nq:], u, p, pd, f, s)
            else:
                self._wsPublisher.publish(t, x[:nq], x[nq:], u)
            time.sleep(dts[i] * factor)

    def getFrameTrajectoryFromSolver(self, solver):
        ps, pds = [], []
        datas = [*solver.problem.runningDatas.tolist(), solver.problem.runningDatas[-1]]
        for i, data in enumerate(datas):
            if self._hasContacts(data):
                pinocchio_data = self._getPinocchioData(data)
                ps.append(self._get_pc(pinocchio_data))
                pds.append(self._get_pdc(pinocchio_data))
        return ps, pds

    def getForceTrajectoryFromSolver(self, solver):
        fs, ss = [], []
        models = [
            *solver.problem.runningModels.tolist(),
            solver.problem.runningModels[-1],
        ]
        datas = [*solver.problem.runningDatas.tolist(), solver.problem.runningDatas[-1]]
        for i, data in enumerate(datas):
            model = models[i]
            if self._hasContacts(data):
                _, contact_data = self._getContactModelAndData(model, data)
                fs.append(self._get_fc(contact_data))
                ss.append(self._get_sc(contact_data))
        return fs, ss

    def setVisibility(self, name, status):
        pass

    def displayFramePoses(self, ps):
        pass

    def displayContactForce(self, f):
        pass

    def displayThrustForce(self, r):
        pass

    def _addForceArrows(self):
        pass

    def _addThrustArrows(self):
        pass

    def _addFrameCurves(self):
        pass

    def _addFrictionCones(self):
        pass

    def _getPinocchioData(self, data):
        if hasattr(data, "differential"):
            if hasattr(data.differential, "multibody"):
                return data.differential.multibody.pinocchio
            elif isinstance(data.differential, StdVec_DiffActionData):
                if hasattr(data.differential[0], "multibody"):
                    return data.differential[0].multibody.pinocchio
        elif isinstance(data, ActionDataImpulseFwdDynamics):
            return data.multibody.pinocchio

    def _get_pc(self, pinocchio_data):
        if len(self.frameTrajNames) == 0:
            return None
        popt = dict()
        for frame in self.frameTrajNames:
            frame_id = int(frame)
            name = self.robot.model.frames[frame_id].name
            pinocchio.updateFramePlacement(self.robot.model, pinocchio_data, frame_id)
            popt[name] = pinocchio_data.oMf[frame_id]
        return popt

    def _get_pdc(self, pinocchio_data):
        if len(self.frameTrajNames) == 0:
            return None
        pdopt = dict()
        for frame in self.frameTrajNames:
            frame_id = int(frame)
            name = self.robot.model.frames[frame_id].name
            v = pinocchio.getFrameVelocity(
                self.robot.model,
                pinocchio_data,
                frame_id,
                pinocchio.ReferenceFrame.LOCAL_WORLD_ALIGNED,
            )
            pdopt[name] = v
        return pdopt

    def _get_fc(self, contact_data):
        import crocoddyl_ros

        if len(self.frameTrajNames) == 0:
            return None
        fc = dict()
        for frame in self.frameTrajNames:
            frame_id = int(frame)
            name = self.robot.model.frames[frame_id].name
            fc[name] = [
                pinocchio.Force.Zero(),
                crocoddyl_ros.ContactType.LOCOMOTION,
                crocoddyl_ros.ContactStatus.SEPARATION,
            ]
        for contact in contact_data:
            force = contact.data().f
            frame_id = contact.data().frame
            name = self.robot.model.frames[frame_id].name
            fc[name] = [
                force,
                crocoddyl_ros.ContactType.LOCOMOTION,
                crocoddyl_ros.ContactStatus.STICKING,
            ]
        return fc

    def _get_sc(self, contact_data):
        sc = dict()
        self.mu = 0.7
        for frame in self.frameTrajNames:
            frame_id = int(frame)
            name = self.robot.model.frames[frame_id].name
            sc[name] = [np.array([0.0, 0.0, 1.0]), self.mu]
        return sc


class CallbackDisplay(CallbackAbstract):
    def __init__(self, display):
        CallbackAbstract.__init__(self)
        self.visualization = display

    def __call__(self, solver):
        if (solver.iter + 1) % self.visualization.freq:
            return
        self.visualization.displayFromSolver(solver)


class CallbackLogger(CallbackAbstract):
    def __init__(self):
        CallbackAbstract.__init__(self)
        self.xs = []
        self.us = []
        self.fs = []
        self.iters = []
        self.costs = []
        self.stops = []
        self.grads = []
        self.pregs = []
        self.dregs = []
        self.steps = []
        self.ffeass = []
        self.hfeass = []

    def __call__(self, solver):
        import copy

        self.xs = copy.copy(solver.xs)
        self.us = copy.copy(solver.us)
        self.fs.append(copy.copy(solver.fs))
        self.iters.append(solver.iter)
        self.costs.append(solver.cost)
        self.stops.append(solver.stoppingCriteria())
        self.grads.append(-solver.expectedImprovement()[1].item())
        self.pregs.append(solver.preg)
        self.dregs.append(solver.dreg)
        self.steps.append(solver.stepLength)
        self.ffeass.append(solver.ffeas)
        self.hfeass.append(solver.hfeas)


def plotOCSolution(xs=None, us=None, figIndex=1, show=True, figTitle=""):
    import matplotlib.pyplot as plt

    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["ps.fonttype"] = 42

    # Getting the state and control trajectories
    if xs is not None:
        xsPlotIdx = 111
        nx = xs[0].shape[0]
        X = [0.0] * nx
        for i in range(nx):
            X[i] = [x[i] for x in xs]
    if us is not None:
        usPlotIdx = 111
        nu = us[0].shape[0]
        U = [0.0] * nu
        for i in range(nu):
            U[i] = [u[i] if u.shape[0] != 0 else 0 for u in us]
    if xs is not None and us is not None:
        xsPlotIdx = 211
        usPlotIdx = 212
    plt.figure(figIndex)

    # Plotting the state trajectories
    if xs is not None:
        plt.subplot(xsPlotIdx)
        [plt.plot(X[i], label="x" + str(i)) for i in range(nx)]
        plt.legend()
        plt.title(figTitle, fontsize=14)

    # Plotting the control commands
    if us is not None:
        plt.subplot(usPlotIdx)
        [plt.plot(U[i], label="u" + str(i)) for i in range(nu)]
        plt.legend()
        plt.xlabel("knots")
    if show:
        plt.show()


def plotFeasibility(ffeass, hfeass, figIndex=1, show=True, figTitle=""):
    import matplotlib.pyplot as plt

    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["ps.fonttype"] = 42
    plt.figure(figIndex, figsize=(6.4, 8))

    # Plotting the feasibility
    plt.ylabel("feasibiltiy")
    plt.plot(ffeass)
    plt.plot(hfeass)
    plt.plot([max(ffeas, hfeas) for ffeas, hfeas in zip(ffeass, hfeass)])
    plt.title(figTitle, fontsize=14)
    plt.xlabel("iteration")
    plt.yscale("log")
    plt.legend(["dynamic", "equality", "total"])
    if show:
        plt.show()


def plotConvergence(
    costs, muLM, muV, gamma, theta, alpha, figIndex=1, show=True, figTitle=""
):
    import matplotlib.pyplot as plt
    import numpy as np

    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["ps.fonttype"] = 42
    plt.figure(figIndex, figsize=(6.4, 8))

    # Plotting the total cost sequence
    plt.subplot(511)
    plt.ylabel("cost")
    plt.plot(costs)
    plt.title(figTitle, fontsize=14)

    # Ploting mu sequences
    plt.subplot(512)
    plt.ylabel("mu")
    plt.plot(muLM, label="LM")
    plt.plot(muV, label="V")
    plt.legend()

    # Plotting the gradient sequence (gamma and theta)
    plt.subplot(513)
    plt.ylabel("gamma")
    plt.plot(gamma)
    plt.subplot(514)
    plt.ylabel("theta")
    plt.plot(theta)

    # Plotting the alpha sequence
    plt.subplot(515)
    plt.ylabel("alpha")
    ind = np.arange(len(alpha))
    plt.bar(ind, alpha)
    plt.xlabel("iteration")
    if show:
        plt.show()


def saveOCSolution(filename, xs, us, ks=None, Ks=None):
    import pickle

    data = {"xs": xs, "us": us, "ks": ks, "Ks": Ks}
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def saveConvergence(filename, costs, muLM, muV, gamma, theta, alpha):
    import pickle

    data = {
        "costs": costs,
        "muLM": muLM,
        "muV": muV,
        "gamma": gamma,
        "theta": theta,
        "alpha": alpha,
    }
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def saveLogfile(filename, log):
    import pickle

    data = {
        "xs": log.xs,
        "us": log.us,
        "fs": log.fs,
        "steps": log.steps,
        "iters": log.iters,
        "costs": log.costs,
        "primal-reg": log.pregs,
        "dual-reg": log.dregs,
        "stops": log.stops,
        "grads": log.grads,
    }
    with open(filename, "wb") as f:
        pickle.dump(data, f)



import numpy as np

EPS = np.finfo(float).eps
NUMDIFF_MODIFIER = 3e4


class NumDiffException(Exception):
    """Raised when the NumDiff values are too high"""

    pass


def assertNumDiff(A, B, threshold):
    """Assert analytical derivatives against NumDiff using the error norm.

    :param A: analytical derivatives
    :param B: NumDiff derivatives
    :param threshold: absolute tolerance
    """
    if not np.allclose(A, B, atol=threshold):
        value = np.linalg.norm(A - B)
        raise NumDiffException(
            f"NumDiff exception, with residual of {value:.4g}, above threshold "
            f"{threshold:.4g}"
        )



"""
Reads csv benchmarks provided as input. Creates a DataFrame for each benchmark, and
prints the table for easy reading.
The benchmarks are stored in the object benches

Depends:
    pandas module (pip install pandas)

How to Run:
    python -i read_csv.py /tmp/bench1.bench /tmp/bench2.bench ...
    e.g.: ipython -i read_csv.py /tmp/Anymal_19DoF.bench /tmp/HyQ_19DoF.bench
    e.g.: ipython -i read_csv.py /tmp/*.bench


Output:
    benches
    benches.bench1 (DataFrame), benches.bench2 (DataFrame)...

DataFrame API:
    value = benches.bench1.loc[fn_name, nthreads, with(out)_cg][bench_parameter]
    example:
    mean_calc_time_withcg_with3threads = benches.bench1.loc["calc",3,True]["mean"]
    max_calc_time_withoutcg_with2threads = benches.bench1.loc["calc",2,False]["max"]

"""

import sys
from os.path import basename, exists, splitext

import pandas as pd

pd.options.display.width = 0


class Benchmarks:
    pass


def parseCsvFile(filename, sep=",", delimiter=None):
    """Inputs a file and ouputs a matrix.
    input: (string) filename
    output: (numpy array) seq
    """
    col_types = dict(
        fn_name=str,
        nthreads=int,
        with_cg=bool,
        mean=float,
        stddev=float,
        max=float,
        min=float,
        mean_per_nodes=float,
        stddev_per_nodes=float,
    )

    seq = pd.read_csv(
        filename,
        header=0,
        sep=sep,
        delim_whitespace=False,
        quoting=2,
        index_col=[0, 1, 2],
        dtype=col_types,
    )

    return seq


benches = Benchmarks()

filenames = sys.argv[1:]
for filename in filenames:
    if not exists(filename):
        continue
    benchname = splitext(basename(filename))[0]
    seq = parseCsvFile(filename)
    setattr(benches, benchname, seq)
    print(benchname)
    print(seq)
    print("********************")



import os
import signal
import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

# In this example test, we will solve the reaching-goal task with the Kinova arm.
# For that, we use the inverse dynamics (with its analytical derivatives) developed
# inside crocoddyl; it is described inside DifferentialActionModelFreeInvDynamics class.
# Finally, we use an Euler sympletic integration scheme.

# First, let's load create the state and actuation models
kinova = example_robot_data.load("kinova")
robot_model = kinova.model
state = crocoddyl.StateMultibody(robot_model)
actuation = crocoddyl.ActuationModelFull(state)
q0 = kinova.model.referenceConfigurations["arm_up"]
x0 = np.concatenate([q0, pinocchio.utils.zero(robot_model.nv)])

# Create a cost model per the running and terminal action model.
nu = state.nv
runningCostModel = crocoddyl.CostModelSum(state, nu)
terminalCostModel = crocoddyl.CostModelSum(state, nu)

# Note that we need to include a cost model (i.e. set of cost functions) in
# order to fully define the action model for our optimal control problem.
# For this particular example, we formulate three running-cost functions:
# goal-tracking cost, state and control regularization; and one terminal-cost:
# goal cost. First, let's create the common cost functions.
framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
    state,
    robot_model.getFrameId("j2s6s200_end_effector"),
    pinocchio.SE3(np.eye(3), np.array([0.6, 0.2, 0.5])),
    nu,
)
uResidual = crocoddyl.ResidualModelJointEffort(state, actuation, nu)
xResidual = crocoddyl.ResidualModelState(state, x0, nu)
goalTrackingCost = crocoddyl.CostModelResidual(state, framePlacementResidual)
xRegCost = crocoddyl.CostModelResidual(state, xResidual)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)

# Then let's added the running and terminal cost functions
runningCostModel.addCost("gripperPose", goalTrackingCost, 1)
runningCostModel.addCost("xReg", xRegCost, 1e-1)
runningCostModel.addCost("uReg", uRegCost, 1e-1)
terminalCostModel.addCost("gripperPose", goalTrackingCost, 1e3)

# Next, we need to create an action model for running and terminal knots. The
# forward dynamics (computed using RNEA) are implemented
# inside DifferentialActionModelFreeInvDynamics.
dt = 1e-2
runningModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeInvDynamics(
        state, actuation, runningCostModel
    ),
    dt,
)
terminalModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeInvDynamics(
        state, actuation, terminalCostModel
    ),
    0.0,
)

# For this optimal control problem, we define 100 knots (or running action
# models) plus a terminal knot
T = 100
problem = crocoddyl.ShootingProblem(x0, [runningModel] * T, terminalModel)

# Creating the solver for this OC problem, defining a logger
solver = crocoddyl.SolverIntro(problem)
if WITHPLOT:
    solver.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    solver.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving it with the solver algorithm
solver.solve()

print(
    "Finally reached = ",
    solver.problem.terminalData.differential.multibody.pinocchio.oMf[
        robot_model.getFrameId("j2s6s200_end_effector")
    ].translation.T,
)

# Plotting the solution and the solver convergence
if WITHPLOT:
    log = solver.getCallbacks()[1]
    crocoddyl.plotOCSolution(
        solver.xs,
        [d.differential.multibody.joint.tau for d in solver.problem.runningDatas],
        figIndex=1,
        show=False,
    )
    crocoddyl.plotConvergence(
        log.costs, log.pregs, log.dregs, log.grads, log.stops, log.steps, figIndex=2
    )

# Visualizing the solution in gepetto-viewer
if WITHDISPLAY:
    try:
        import gepetto

        cameraTF = [2.0, 2.68, 0.54, 0.2, 0.62, 0.72, 0.22]
        gepetto.corbaserver.Client()
        display = crocoddyl.GepettoDisplay(kinova, 4, 4, cameraTF, floor=False)
    except Exception:
        display = crocoddyl.MeshcatDisplay(kinova)

    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)



import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl
from crocoddyl.utils.biped import SimpleBipedGaitProblem

T = int(sys.argv[1]) if (len(sys.argv) > 1) else int(5e3)  # number of trials
MAXITER = 1
GAIT = "walking"  # 55 nodes


def createProblem(gait_phase):
    robot_model = example_robot_data.load("talos_legs").model
    rightFoot, leftFoot = "right_sole_link", "left_sole_link"
    gait = SimpleBipedGaitProblem(robot_model, rightFoot, leftFoot)
    q0 = robot_model.referenceConfigurations["half_sitting"].copy()
    v0 = pinocchio.utils.zero(robot_model.nv)
    x0 = np.concatenate([q0, v0])

    type_of_gait = next(iter(gait_phase.keys()))
    value = gait_phase[type_of_gait]
    if type_of_gait == "walking":
        # Creating a walking problem
        problem = gait.createWalkingProblem(
            x0,
            value["stepLength"],
            value["stepHeight"],
            value["timeStep"],
            value["stepKnots"],
            value["supportKnots"],
        )

    xs = [robot_model.defaultState] * (len(problem.runningModels) + 1)
    us = [
        m.quasiStatic(d, robot_model.defaultState)
        for m, d in list(zip(problem.runningModels, problem.runningDatas))
    ]
    return xs, us, problem


def runDDPSolveBenchmark(xs, us, problem):
    solver = crocoddyl.SolverFDDP(problem)

    duration = []
    for _ in range(T):
        c_start = time.time()
        solver.solve(xs, us, MAXITER, False, 0.1)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcBenchmark(xs, us, problem):
    duration = []
    for _ in range(T):
        c_start = time.time()
        problem.calc(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcDiffBenchmark(xs, us, problem):
    duration = []
    for _ in range(T):
        c_start = time.time()
        problem.calcDiff(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


# Setting up all tasks
if GAIT == "walking":
    GAITPHASE = {
        "walking": {
            "stepLength": 0.6,
            "stepHeight": 0.1,
            "timeStep": 0.0375,
            "stepKnots": 25,
            "supportKnots": 1,
        }
    }

xs, us, problem = createProblem(GAITPHASE)
print("NQ:", problem.terminalModel.state.nq)
print("Number of nodes:", problem.T)
avrg_dur, min_dur, max_dur = runDDPSolveBenchmark(xs, us, problem)
print(f"  FDDP.solve [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcBenchmark(xs, us, problem)
print(f"  ShootingProblem.calc [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcDiffBenchmark(xs, us, problem)
print(f"  ShootingProblem.calcDiff [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")



import numpy as np

import crocoddyl


class CostModelDoublePendulum(crocoddyl.CostModelAbstract):
    def __init__(self, state, activation, nu):
        activation = (
            activation
            if activation is not None
            else crocoddyl.ActivationModelQuad(state.ndx)
        )
        crocoddyl.CostModelAbstract.__init__(self, state, activation, nu=nu)

    def calc(self, data, x, u=None):
        c1, c2 = np.cos(x[0]), np.cos(x[1])
        s1, s2 = np.sin(x[0]), np.sin(x[1])
        data.residual.r[:] = np.array([s1, s2, 1 - c1, 1 - c2, x[2], x[3]])
        self.activation.calc(data.activation, data.residual.r)
        data.cost = data.activation.a_value

    def calcDiff(self, data, x, u=None):
        c1, c2 = np.cos(x[0]), np.cos(x[1])
        s1, s2 = np.sin(x[0]), np.sin(x[1])

        self.activation.calcDiff(data.activation, data.residual.r)

        data.residual.Rx[:2, :2] = np.diag([c1, c2])
        data.residual.Rx[2:4, :2] = np.diag([s1, s2])
        data.residual.Rx[4:6, 2:4] = np.diag([1, 1])
        data.Lx[:] = np.dot(data.residual.Rx.T, data.activation.Ar)

        data.Rxx[:2, :2] = np.diag([c1**2 - s1**2, c2**2 - s2**2])
        data.Rxx[2:4, :2] = np.diag([s1**2 + (1 - c1) * c1, s2**2 + (1 - c2) * c2])
        data.Rxx[4:6, 2:4] = np.diag([1, 1])
        data.Lxx[:, :] = np.diag(np.dot(data.Rxx.T, np.diag(data.activation.Arr)))

    def createData(self, collector):
        data = CostDataDoublePendulum(self, collector)
        return data


class CostDataDoublePendulum(crocoddyl.CostDataAbstract):
    def __init__(self, model, collector):
        crocoddyl.CostDataAbstract.__init__(self, model, collector)
        self.Rxx = np.zeros((6, 4))


class ActuationModelDoublePendulum(crocoddyl.ActuationModelAbstract):
    def __init__(self, state, actLink):
        crocoddyl.ActuationModelAbstract.__init__(self, state, 1)
        self.nv = state.nv
        self.actLink = actLink

    def calc(self, data, x, u):
        data.tau[:] = data.dtau_du * u

    def calcDiff(self, data, x, u):
        pass

    def commands(self, data, x, tau):
        if self.actLink == 1:
            data.u[:] = tau[0]
        else:
            data.u[:] = tau[1]

    def torqueTransform(self, data, x, tau):
        pass

    def createData(self):
        data = ActuationDataDoublePendulum(self)
        return data


class ActuationDataDoublePendulum(crocoddyl.ActuationDataAbstract):
    def __init__(self, model):
        crocoddyl.ActuationDataAbstract.__init__(self, model)
        if model.actLink == 1:
            self.dtau_du[0] = 1.0
            self.tau_set = [True, False]
            self.Mtau[0] = 1.0
        else:
            self.dtau_du[1] = 1.0
            self.tau_set = [False, True]
            self.Mtau[1] = 1.0



from math import cos, sin

import numpy as np
from matplotlib import animation
from matplotlib import pyplot as plt


def animateCartpole(xs, sleep=50, show=False):
    print("processing the animation ... ")
    cart_size = 1.0
    pole_length = 5.0
    fig = plt.figure()
    ax = plt.axes(xlim=(-8, 8), ylim=(-6, 6))
    patch = plt.Rectangle((0.0, 0.0), cart_size, cart_size, fc="b")
    (line,) = ax.plot([], [], "k-", lw=2)
    time_text = ax.text(0.02, 0.95, "", transform=ax.transAxes)

    def init():
        ax.add_patch(patch)
        line.set_data([], [])
        time_text.set_text("")
        return patch, line, time_text

    def animate(i):
        x_cart = xs[i][0]
        y_cart = 0.0
        theta = xs[i][1]
        patch.set_xy([x_cart - cart_size / 2, y_cart - cart_size / 2])
        x_pole = np.cumsum([x_cart, -pole_length * sin(theta)])
        y_pole = np.cumsum([y_cart, pole_length * cos(theta)])
        line.set_data(x_pole, y_pole)
        time = i * sleep / 1000.0
        time_text.set_text(f"time = {time:.1f} sec")
        return patch, line, time_text

    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=len(xs), interval=sleep, blit=True
    )
    print("... processing done")
    if show:
        plt.show()
    return anim



import numpy as np
import pinocchio

# from crocoddyl.utils import EPS
from pinocchio.utils import zero

EPS = np.finfo(float).eps
NUMDIFF_MODIFIER = 1e4


class NumDiffException(Exception):
    """Raised when the NumDiff values are too high"""

    pass


def assertNumDiff(A, B, threshold):
    """Assert analytical derivatives against NumDiff using the error norm.

    :param A: analytical derivatives
    :param B: NumDiff derivatives
    :param threshold: absolute tolerance
    """
    if not np.allclose(A, B, atol=threshold):
        value = np.linalg.norm(A - B)
        raise NumDiffException(
            f"NumDiff exception, with residual of {value:.4g}, above threshold "
            f"{threshold:.4g}"
        )


def df_dx(func, x, h=np.sqrt(2 * EPS)):
    """Perform df/dx by num_diff.
    :params func: function to differentiate f : np.matrix -> np.matrix
    :params x: value at which f is differentiated. type np.matrix
    :params h: eps

    :returns df/dx
    """
    dx = zero(x.size)
    f0 = func(x)
    res = zero([len(f0), x.size])
    for ix in range(x.size):
        dx[ix] = h
        res[:, ix] = (func(x + dx) - f0) / h
        dx[ix] = 0
    return res


def df_dq(model, func, q, h=np.sqrt(2 * EPS)):
    """Perform df/dq by num_diff. q is in the lie manifold.
    :params func: function to differentiate f : np.matrix -> np.matrix
    :params q: configuration value at which f is differentiated. type np.matrix
    :params h: eps

    :returns df/dq
    """
    dq = zero(model.nv)
    f0 = func(q)
    res = zero([len(f0), model.nv])
    for iq in range(model.nv):
        dq[iq] = h
        res[:, iq] = (func(pinocchio.integrate(model, q, dq)) - f0) / h
        dq[iq] = 0
    return res



import sys
import unittest

import example_robot_data
import numpy as np
import pinocchio
from factory import FreeFloatingActuationDerived, FullActuationDerived

import crocoddyl


class ActuationModelAbstractTestCase(unittest.TestCase):
    STATE = None
    ACTUATION = None
    ACTUATION_DER = None

    def setUp(self):
        self.x = self.STATE.rand()
        self.u = pinocchio.utils.rand(self.ACTUATION.nu)
        self.DATA = self.ACTUATION.createData()
        self.DATA_DER = self.ACTUATION_DER.createData()

    def test_calc(self):
        # Run calc for both action models
        self.ACTUATION.calc(self.DATA, self.x, self.u)
        self.ACTUATION_DER.calc(self.DATA_DER, self.x, self.u)
        # Checking the actuation signal
        self.assertTrue(
            np.allclose(self.DATA.tau, self.DATA_DER.tau, atol=1e-9),
            "Wrong actuation signal.",
        )

    def test_calcDiff(self):
        # Run calcDiff for both action models
        self.ACTUATION.calcDiff(self.DATA, self.x, self.u)
        self.ACTUATION_DER.calcDiff(self.DATA_DER, self.x, self.u)
        # Checking the Jacobians of the actuation model
        self.assertTrue(
            np.allclose(self.DATA.dtau_dx, self.DATA_DER.dtau_dx, atol=1e-9),
            "Wrong dtau_dx.",
        )
        self.assertTrue(
            np.allclose(self.DATA.dtau_du, self.DATA_DER.dtau_du, atol=1e-9),
            "Wrong dtau_du.",
        )


class TalosArmFullActuationTest(ActuationModelAbstractTestCase):
    STATE = crocoddyl.StateMultibody(example_robot_data.load("talos_arm").model)

    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    ACTUATION_DER = FullActuationDerived(STATE)


class HyQFloatingBaseActuationTest(ActuationModelAbstractTestCase):
    STATE = crocoddyl.StateMultibody(example_robot_data.load("hyq").model)

    ACTUATION = crocoddyl.ActuationModelFloatingBase(STATE)
    ACTUATION_DER = FreeFloatingActuationDerived(STATE)


class TalosFloatingBaseActuationTest(ActuationModelAbstractTestCase):
    STATE = crocoddyl.StateMultibody(example_robot_data.load("talos").model)

    ACTUATION = crocoddyl.ActuationModelFloatingBase(STATE)
    ACTUATION_DER = FreeFloatingActuationDerived(STATE)


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        TalosArmFullActuationTest,
        HyQFloatingBaseActuationTest,
        TalosFloatingBaseActuationTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import sys
import unittest

import numpy as np
import pinocchio
from factory import SquashingSmoothSatDerived

import crocoddyl


class SquashingModelAbstractTestCase(unittest.TestCase):
    SQUASHING = None
    SQUASHING_DER = None

    def setUp(self):
        self.s = pinocchio.utils.rand(self.SQUASHING.ns)
        self.DATA = self.SQUASHING.createData()
        self.DATA_DER = self.SQUASHING_DER.createData()

    def test_calc(self):
        # Run calc for both squashing functions
        self.SQUASHING.calc(self.DATA, self.s)
        self.SQUASHING_DER.calc(self.DATA_DER, self.s)
        # Checking the squashing signal
        self.assertTrue(
            np.allclose(self.DATA.u, self.DATA_DER.u, atol=1e-9),
            "Wrong squashing signal.",
        )

    def test_calcDiff(self):
        # Run calcDiff for both squashing functions
        self.SQUASHING.calcDiff(self.DATA, self.s)
        self.SQUASHING_DER.calcDiff(self.DATA_DER, self.s)
        # Checking teh Jacobians of the squashing function
        self.assertTrue(
            np.allclose(self.DATA.du_ds, self.DATA_DER.du_ds, atol=1e-9), "Wrong ds_du."
        )


class SmoothSatSquashingTest(SquashingModelAbstractTestCase):
    NS = 4
    U_UB = pinocchio.utils.zero(NS)
    U_LB = pinocchio.utils.zero(NS)
    U_UB.fill(10)
    U_LB.fill(0)
    SQUASHING = crocoddyl.SquashingModelSmoothSat(U_LB, U_UB, NS)
    SQUASHING_DER = SquashingSmoothSatDerived(U_LB, U_UB, NS)


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [SmoothSatSquashingTest]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



# flake8: noqa: E203
# ref. https://github.com/PyCQA/pycodestyle/issues/373, remove this for ruff
import os
import signal
import sys
import time

import example_robot_data
import numpy as np

import crocoddyl
from crocoddyl.utils.pendulum import (
    ActuationModelDoublePendulum,
    CostModelDoublePendulum,
)

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Loading the double pendulum model
pendulum = example_robot_data.load("double_pendulum")
model = pendulum.model

state = crocoddyl.StateMultibody(model)
actuation = ActuationModelDoublePendulum(state, actLink=1)

nu = state.nv
runningCostModel = crocoddyl.CostModelSum(state, nu)
terminalCostModel = crocoddyl.CostModelSum(state, nu)

xResidual = crocoddyl.ResidualModelState(state, state.zero(), nu)
xActivation = crocoddyl.ActivationModelQuad(state.ndx)
uResidual = crocoddyl.ResidualModelJointEffort(
    state, actuation, np.zeros(actuation.nu), nu, False
)
xRegCost = crocoddyl.CostModelResidual(state, xActivation, xResidual)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
xPendCost = CostModelDoublePendulum(
    state, crocoddyl.ActivationModelWeightedQuad(np.array([1.0] * 4 + [0.1] * 2)), nu
)

dt = 1e-2

runningCostModel.addCost("uReg", uRegCost, 1e-4 / dt)
runningCostModel.addCost("xGoal", xPendCost, 1e-5 / dt)
terminalCostModel.addCost("xGoal", xPendCost, 100.0)

runningModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeInvDynamics(
        state, actuation, runningCostModel
    ),
    dt,
)
terminalModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeInvDynamics(
        state, actuation, terminalCostModel
    ),
    dt,
)

# Creating the shooting problem and the solver
T = 100
x0 = np.array([3.14, 0.0, 0.0, 0.0])
problem = crocoddyl.ShootingProblem(x0, [runningModel] * T, terminalModel)
solver = crocoddyl.SolverIntro(problem)
if WITHPLOT:
    solver.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    solver.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving the problem with the solver
solver.solve()

# Plotting the entire motion
if WITHPLOT:
    log = solver.getCallbacks()[1]
    crocoddyl.plotOCSolution(
        log.xs, [u[state.nv :] for u in log.us], figIndex=1, show=False
    )
    crocoddyl.plotConvergence(
        log.costs, log.pregs, log.dregs, log.grads, log.stops, log.steps, figIndex=2
    )

# Display the entire motion
if WITHDISPLAY:
    try:
        import gepetto

        gepetto.corbaserver.Client()
        cameraTF = [1.4, 0.0, 0.2, 0.5, 0.5, 0.5, 0.5]
        display = crocoddyl.GepettoDisplay(pendulum, 4, 4, cameraTF, floor=False)
    except Exception:
        display = crocoddyl.MeshcatDisplay(pendulum)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)



# -*- python -*-
# Copyright (C) 2008-2018 LAAS-CNRS, JRL AIST-CNRS.
# Author: Joseph Mirabel
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from omniidl_be.python import output_inline
from omniidl_be.python import run as run_parent
from omniidl import idlvisitor, idlast, idltype


def _rreplace(s, old, new, occurrence):
    li = s.rsplit(old, occurrence)
    return new.join(li)


class CommentToConstVisitor(idlvisitor.AstVisitor):
    def __init__(self):
        import re

        self.commentStart = re.compile(r"^//+ ?")

    def _commentToConst(self, node, comments):
        import re

        texts = []
        line = node.line()
        # TODO here, we only use the comments before.
        comment = node
        for c in reversed(comments):
            if line - c.line() != 1:
                break
            line = c.line()
            comment = c
            text = c.text()
            texts.append(re.sub(self.commentStart, "", text))
        if len(texts) == 0:
            if isinstance(node, idlast.Attribute):
                for i in node.identifiers():
                    print(i + " documentation may be ill-formed.")
            else:
                print(node.identifier() + " documentation may be ill-formed.")
        texts.reverse()
        # Extract the prototype of the function
        try:
            if isinstance(node, idlast.Operation):
                with open(node.file()) as fp:
                    for _ in range(node.line() - 1):
                        fp.readline()
                    line = fp.readline()
                    texts.append("\nPrototype:\t" + line)
                    while line != "":
                        if ";" in line:
                            break
                        line = fp.readline()
                        texts.append("           \t" + line)
        except Exception:
            pass
        if len(texts) == 0:
            return None
        text = "".join(texts)
        id = node.identifier() + "__doc__"
        return idlast.Const(
            comment.file(),
            comment.line(),
            node.mainFile(),
            node.pragmas(),
            [],  # comments
            id,  # identifier
            node.scopedName()[:-1]
            + [
                id,
            ],  # scopedName
            _rreplace(node.repoId(), node.identifier(), id, 1),  # repoId
            idltype.String(0),  # constType, 0 means unbounded
            idltype.tk_string,  # constKind
            text,  # value
        )

    def _addDoc(self, parent, node):
        # Commented in order to add the prototype to the documentation.
        # if len(node.comments()) > 0:
        if True:
            const = self._commentToConst(node, node.comments())
            if const is None:
                return
            if isinstance(parent, idlast.Module):
                parent._Module__definitions.append(const)
            elif isinstance(parent, idlast.Interface):
                parent._Interface__declarations.append(const)
                parent._Interface__contents.append(const)
            elif isinstance(parent, idlast.AST):
                parent._AST__declarations.append(const)
            else:
                print("%s:%s: warning: doc ignored" % (node.file(), node.line()))

    def visitAST(self, node):
        for n in node.declarations():
            if not output_inline and not n.mainFile():
                continue

            if isinstance(n, idlast.Module) or isinstance(n, idlast.Interface):
                self._addDoc(node, n)
                n.accept(self)

    def visitModule(self, node):
        for n in node.definitions():
            if not output_inline and not n.mainFile():
                continue

            if isinstance(n, idlast.Module) or isinstance(n, idlast.Interface):
                self._addDoc(node, n)
                n.accept(self)

    def visitInterface(self, node):
        for c in node.callables():
            self._addDoc(node, c)


def run(tree, args):
    ccv = CommentToConstVisitor()
    tree.accept(ccv)

    run_parent(tree, args)



# flake8: noqa: F821
def cartpole_analytical_derivatives(model, data, x, u=None):
    if u is None:
        u = model.unone

    # Getting the state and control variables
    y, th, ydot, thdot = x[0].item(), x[1].item(), x[2].item(), x[3].item()
    f = u[0].item()

    # Shortname for system parameters
    m1, m2, lcart, g = model.m1, model.m2, model.l, model.g
    s, c = np.sin(th), np.cos(th)
    m = m1 + m2
    mu = m1 + m2 * s**2
    w = model.costWeights

    # derivative of xddot by x, theta, xdot, thetadot
    # derivative of thddot by x, theta, xdot, thetadot
    data.Fx[:, :] = np.array(
        [
            [
                0.0,
                (m2 * g * c * c - m2 * g * s * s - m2 * lcart * c * thdot) / mu,
                0.0,
                -m2 * lcart * s / mu,
            ],
            [
                0.0,
                (
                    (-s * f / lcart)
                    + (m * g * c / lcart)
                    - (m2 * c * c * thdot**2)
                    + (m2 * s * s * thdot**2)
                )
                / mu,
                0.0,
                -2 * m2 * c * s * thdot,
            ],
        ]
    )
    # derivative of xddot and thddot by f
    data.Fu[:] = np.array([1 / mu, c / (lcart * mu)])
    # first derivative of data.cost by x, theta, xdot, thetadot
    data.Lx[:] = np.array(
        [
            y * w[2] ** 2,
            s * ((w[0] ** 2 - w[1] ** 2) * c + w[1] ** 2),
            ydot * w[3] ** 2,
            thdot * w[4] ** 2,
        ]
    )
    # first derivative of data.cost by f
    data.Lu[:] = np.array([f * w[5] ** 2])
    # second derivative of data.cost by x, theta, xdot, thetadot
    data.Lxx[:] = np.array(
        [
            w[2] ** 2,
            w[0] ** 2 * (c**2 - s**2) + w[1] ** 2 * (s**2 - c**2 + c),
            w[3] ** 2,
            w[4] ** 2,
        ]
    )
    # second derivative of data.cost by f
    data.Luu[:] = np.array([w[5] ** 2])



import sys
import unittest
from random import randint

import example_robot_data
import numpy as np
import pinocchio
from factory import DifferentialFreeFwdDynamicsModelDerived, UnicycleModelDerived

import crocoddyl


class ShootingProblemTestCase(unittest.TestCase):
    MODEL = None
    MODEL_DER = None

    def setUp(self):
        self.T = randint(1, 101)
        state = self.MODEL.state
        self.xs = []
        self.us = []
        self.xs.append(state.rand())
        rng = np.random.default_rng()
        for _ in range(self.T):
            self.xs.append(state.rand())
            self.us.append(rng.random(self.MODEL.nu))
        self.PROBLEM = crocoddyl.ShootingProblem(
            self.xs[0], [self.MODEL] * self.T, self.MODEL
        )
        self.PROBLEM_DER = crocoddyl.ShootingProblem(
            self.xs[0], [self.MODEL_DER] * self.T, self.MODEL_DER
        )

    def test_number_of_nodes(self):
        self.assertEqual(self.T, self.PROBLEM.T, "Wrong number of nodes")

    def test_calc(self):
        # Running calc functions
        cost = self.PROBLEM.calc(self.xs, self.us)
        costDer = self.PROBLEM_DER.calc(self.xs, self.us)
        self.assertAlmostEqual(cost, costDer, 10, "Wrong cost value")
        for d1, d2 in zip(self.PROBLEM.runningDatas, self.PROBLEM_DER.runningDatas):
            self.assertTrue(
                np.allclose(d1.xnext, d2.xnext, atol=1e-9), "Next state doesn't match."
            )

    def test_calcDiff(self):
        # Running calc functions
        cost = self.PROBLEM.calc(self.xs, self.us)
        costDer = self.PROBLEM_DER.calc(self.xs, self.us)

        cost = self.PROBLEM.calcDiff(self.xs, self.us)
        costDer = self.PROBLEM_DER.calcDiff(self.xs, self.us)
        self.assertAlmostEqual(cost, costDer, 10, "Wrong cost value")
        for d1, d2 in zip(self.PROBLEM.runningDatas, self.PROBLEM_DER.runningDatas):
            self.assertTrue(
                np.allclose(d1.xnext, d2.xnext, atol=1e-9), "Next state doesn't match."
            )
            self.assertTrue(np.allclose(d1.Lx, d2.Lx, atol=1e-9), "Lx doesn't match.")
            self.assertTrue(np.allclose(d1.Lu, d2.Lu, atol=1e-9), "Lu doesn't match.")
            self.assertTrue(
                np.allclose(d1.Lxx, d2.Lxx, atol=1e-9), "Lxx doesn't match."
            )
            self.assertTrue(
                np.allclose(d1.Lxu, d2.Lxu, atol=1e-9), "Lxu doesn't match."
            )
            self.assertTrue(
                np.allclose(d1.Luu, d2.Luu, atol=1e-9), "Luu doesn't match."
            )
            self.assertTrue(np.allclose(d1.Fx, d2.Fx, atol=1e-9), "Fx doesn't match.")
            self.assertTrue(np.allclose(d1.Fu, d2.Fu, atol=1e-9), "Fu doesn't match.")

    def test_rollout(self):
        xs = self.PROBLEM.rollout(self.us)
        xsDer = self.PROBLEM_DER.rollout(self.us)
        for x1, x2 in zip(xs, xsDer):
            self.assertTrue(
                np.allclose(x1, x2, atol=1e-9), "The rollout state doesn't match."
            )


class UnicycleShootingTest(ShootingProblemTestCase):
    MODEL = crocoddyl.ActionModelUnicycle()
    MODEL_DER = UnicycleModelDerived()


class TalosArmShootingTest(ShootingProblemTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-3,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    DIFF_MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    DIFF_MODEL_DER = DifferentialFreeFwdDynamicsModelDerived(STATE, ACTUATION, COST_SUM)
    MODEL = crocoddyl.IntegratedActionModelEuler(DIFF_MODEL, 1e-3)
    MODEL_DER = crocoddyl.IntegratedActionModelEuler(DIFF_MODEL_DER, 1e-3)


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [UnicycleShootingTest, TalosArmShootingTest]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import matplotlib.pyplot as plt
import numpy as np


def plotUnicycle(x):
    sc, delta = 0.1, 0.1
    a, b, th = x[0], x[1], x[2]
    c, s = np.cos(th), np.sin(th)
    refs = [
        plt.arrow(
            a - sc / 2 * c - delta * s,
            b - sc / 2 * s + delta * c,
            c * sc,
            s * sc,
            head_width=0.05,
        ),
        plt.arrow(
            a - sc / 2 * c + delta * s,
            b - sc / 2 * s - delta * c,
            c * sc,
            s * sc,
            head_width=0.05,
        ),
    ]
    return refs


def plotUnicycleSolution(xs, figIndex=1, show=True):
    plt.figure(figIndex, figsize=(6.4, 6.4))
    for x in xs:
        plotUnicycle(x)
    plt.axis([-2, 2.0, -2.0, 2.0])
    if show:
        plt.show()



# flake8: noqa: E203
# ref. https://github.com/PyCQA/pycodestyle/issues/373, remove this for ruff
import numpy as np
import pinocchio

import crocoddyl


class SimpleQuadrupedalGaitProblem:
    """Build simple quadrupedal locomotion problems.

    This class aims to build simple locomotion problems used in the examples of
    Crocoddyl.
    The scope of this class is purely for academic reasons, and it does not aim to be
    used in any robotics application.
    We also do not consider it as part of the API, so changes in this class will not
    pass through a strict process of deprecation.
    Thus, we advice any user to DO NOT develop their application based on this class.
    """

    def __init__(
        self,
        rmodel,
        lfFoot,
        rfFoot,
        lhFoot,
        rhFoot,
        integrator="euler",
        control="zero",
        fwddyn=True,
    ):
        """Construct quadrupedal-gait problem.

        :param rmodel: robot model
        :param lfFoot: name of the left-front foot
        :param rfFoot: name of the right-front foot
        :param lhFoot: name of the left-hind foot
        :param rhFoot: name of the right-hind foot
        :param integrator: type of the integrator
            (options are: 'euler', and 'rk4')
        :param control: type of control parametrization
            (options are: 'zero', 'one', and 'rk4')
        :param fwddyn: True for forward-dynamics and False for inverse-dynamics
            formulations
        """
        self.rmodel = rmodel
        self.rdata = rmodel.createData()
        self.state = crocoddyl.StateMultibody(self.rmodel)
        self.actuation = crocoddyl.ActuationModelFloatingBase(self.state)
        # Getting the frame id for all the legs
        self.lfFootId = self.rmodel.getFrameId(lfFoot)
        self.rfFootId = self.rmodel.getFrameId(rfFoot)
        self.lhFootId = self.rmodel.getFrameId(lhFoot)
        self.rhFootId = self.rmodel.getFrameId(rhFoot)
        self._integrator = integrator
        self._control = control
        self._fwddyn = fwddyn

        # Defining default state
        q0 = self.rmodel.referenceConfigurations["standing"]
        self.rmodel.defaultState = np.concatenate([q0, np.zeros(self.rmodel.nv)])
        self.firstStep = True
        # Defining the friction coefficient and normal
        self.mu = 0.7
        self.Rsurf = np.eye(3)

    def createCoMProblem(self, x0, comGoTo, timeStep, numKnots):
        """Create a shooting problem for a CoM forward/backward task.

        :param x0: initial state
        :param comGoTo: initial CoM motion
        :param timeStep: step time for each knot
        :param numKnots: number of knots per each phase
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        com0 = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)

        # Defining the action models along the time instances
        comModels = []

        # Creating the action model for the CoM task
        comForwardModels = [
            self.createSwingFootModel(
                timeStep, [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId]
            )
            for k in range(numKnots)
        ]
        comForwardTermModel = self.createSwingFootModel(
            timeStep,
            [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            com0 + np.array([comGoTo, 0.0, 0.0]),
        )
        comForwardTermModel.differential.costs.costs["comTrack"].weight = 1e6

        comBackwardModels = [
            self.createSwingFootModel(
                timeStep, [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId]
            )
            for k in range(numKnots)
        ]
        comBackwardTermModel = self.createSwingFootModel(
            timeStep,
            [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            com0 + np.array([-comGoTo, 0.0, 0.0]),
        )
        comBackwardTermModel.differential.costs.costs["comTrack"].weight = 1e6

        # Adding the CoM tasks
        comModels += [*comForwardModels, comForwardTermModel]
        comModels += [*comBackwardModels, comBackwardTermModel]

        # Defining the shooting problem
        problem = crocoddyl.ShootingProblem(x0, comModels[:-1], comModels[-1])
        return problem

    def createCoMGoalProblem(self, x0, comGoTo, timeStep, numKnots):
        """Create a shooting problem for a CoM position goal task.

        :param x0: initial state
        :param comGoTo: CoM position change target
        :param timeStep: step time for each knot
        :param numKnots: number of knots per each phase
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.state.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        com0 = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)

        # Defining the action models along the time instances
        comModels = []

        # Creating the action model for the CoM task
        comForwardModels = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            )
            for k in range(numKnots)
        ]
        comForwardTermModel = self.createSwingFootModel(
            timeStep,
            [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            com0 + np.array([comGoTo, 0.0, 0.0]),
        )
        comForwardTermModel.differential.costs.costs["comTrack"].weight = 1e6

        # Adding the CoM tasks
        comModels += [*comForwardModels, comForwardTermModel]

        # Defining the shooting problem
        problem = crocoddyl.ShootingProblem(x0, comModels[:-1], comModels[-1])
        return problem

    def createWalkingProblem(
        self, x0, stepLength, stepHeight, timeStep, stepKnots, supportKnots
    ):
        """Create a shooting problem for a simple walking gait.

        :param x0: initial state
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: step time for each knot
        :param stepKnots: number of knots for step phases
        :param supportKnots: number of knots for double support phases
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.state.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfFootId].translation
        rhFootPos0 = self.rdata.oMf[self.rhFootId].translation
        lfFootPos0 = self.rdata.oMf[self.lfFootId].translation
        lhFootPos0 = self.rdata.oMf[self.lhFootId].translation
        comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2].item()

        # Defining the action models along the time instances
        loco3dModel = []
        doubleSupport = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            )
            for k in range(supportKnots)
        ]
        if self.firstStep is True:
            rhStep = self.createFootstepModels(
                comRef,
                [rhFootPos0],
                0.5 * stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.rfFootId, self.lhFootId],
                [self.rhFootId],
            )
            rfStep = self.createFootstepModels(
                comRef,
                [rfFootPos0],
                0.5 * stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.lhFootId, self.rhFootId],
                [self.rfFootId],
            )
            self.firstStep = False
        else:
            rhStep = self.createFootstepModels(
                comRef,
                [rhFootPos0],
                stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.rfFootId, self.lhFootId],
                [self.rhFootId],
            )
            rfStep = self.createFootstepModels(
                comRef,
                [rfFootPos0],
                stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.lhFootId, self.rhFootId],
                [self.rfFootId],
            )
        lhStep = self.createFootstepModels(
            comRef,
            [lhFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.lfFootId, self.rfFootId, self.rhFootId],
            [self.lhFootId],
        )
        lfStep = self.createFootstepModels(
            comRef,
            [lfFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.rfFootId, self.lhFootId, self.rhFootId],
            [self.lfFootId],
        )
        loco3dModel += doubleSupport + rhStep + rfStep
        loco3dModel += doubleSupport + lhStep + lfStep + [doubleSupport[0]]

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createTrottingProblem(
        self, x0, stepLength, stepHeight, timeStep, stepKnots, supportKnots
    ):
        """Create a shooting problem for a simple trotting gait.

        :param x0: initial state
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: step time for each knot
        :param stepKnots: number of knots for step phases
        :param supportKnots: number of knots for double support phases
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfFootId].translation
        rhFootPos0 = self.rdata.oMf[self.rhFootId].translation
        lfFootPos0 = self.rdata.oMf[self.lfFootId].translation
        lhFootPos0 = self.rdata.oMf[self.lhFootId].translation
        comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2].item()

        # Defining the action models along the time instances
        loco3dModel = []
        doubleSupport = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            )
            for k in range(supportKnots)
        ]
        if self.firstStep is True:
            rflhStep = self.createFootstepModels(
                comRef,
                [rfFootPos0, lhFootPos0],
                0.5 * stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.rhFootId],
                [self.rfFootId, self.lhFootId],
            )
            self.firstStep = False
        else:
            rflhStep = self.createFootstepModels(
                comRef,
                [rfFootPos0, lhFootPos0],
                stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.rhFootId],
                [self.rfFootId, self.lhFootId],
            )
        lfrhStep = self.createFootstepModels(
            comRef,
            [lfFootPos0, rhFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.rfFootId, self.lhFootId],
            [self.lfFootId, self.rhFootId],
        )

        loco3dModel += doubleSupport + rflhStep
        loco3dModel += doubleSupport + lfrhStep + [doubleSupport[0]]

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createPacingProblem(
        self, x0, stepLength, stepHeight, timeStep, stepKnots, supportKnots
    ):
        """Create a shooting problem for a simple pacing gait.

        :param x0: initial state
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: step time for each knot
        :param stepKnots: number of knots for step phases
        :param supportKnots: number of knots for double support phases
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfFootId].translation
        rhFootPos0 = self.rdata.oMf[self.rhFootId].translation
        lfFootPos0 = self.rdata.oMf[self.lfFootId].translation
        lhFootPos0 = self.rdata.oMf[self.lhFootId].translation
        comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2].item()

        # Defining the action models along the time instances
        loco3dModel = []
        doubleSupport = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            )
            for k in range(supportKnots)
        ]
        if self.firstStep is True:
            rightSteps = self.createFootstepModels(
                comRef,
                [rfFootPos0, rhFootPos0],
                0.5 * stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.lhFootId],
                [self.rfFootId, self.rhFootId],
            )
            self.firstStep = False
        else:
            rightSteps = self.createFootstepModels(
                comRef,
                [rfFootPos0, rhFootPos0],
                stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfFootId, self.lhFootId],
                [self.rfFootId, self.rhFootId],
            )
        leftSteps = self.createFootstepModels(
            comRef,
            [lfFootPos0, lhFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.rfFootId, self.rhFootId],
            [self.lfFootId, self.lhFootId],
        )

        loco3dModel += doubleSupport + rightSteps
        loco3dModel += doubleSupport + leftSteps + [doubleSupport[0]]

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createBoundingProblem(
        self, x0, stepLength, stepHeight, timeStep, stepKnots, supportKnots
    ):
        """Create a shooting problem for a simple bounding gait.

        :param x0: initial state
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: step time for each knot
        :param stepKnots: number of knots for step phases
        :param supportKnots: number of knots for double support phases
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfFootId].translation
        rhFootPos0 = self.rdata.oMf[self.rhFootId].translation
        lfFootPos0 = self.rdata.oMf[self.lfFootId].translation
        lhFootPos0 = self.rdata.oMf[self.lhFootId].translation
        comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2].item()

        # Defining the action models along the time instances
        loco3dModel = []
        doubleSupport = [
            self.createSwingFootModel(
                timeStep, [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId]
            )
            for k in range(supportKnots)
        ]
        hindSteps = self.createFootstepModels(
            comRef,
            [lfFootPos0, rfFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.lhFootId, self.rhFootId],
            [self.lfFootId, self.rfFootId],
        )
        frontSteps = self.createFootstepModels(
            comRef,
            [lhFootPos0, rhFootPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.lfFootId, self.rfFootId],
            [self.lhFootId, self.rhFootId],
        )

        loco3dModel += doubleSupport + hindSteps
        loco3dModel += doubleSupport + frontSteps + [doubleSupport[0]]

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createJumpingProblem(
        self, x0, jumpHeight, jumpLength, timeStep, groundKnots, flyingKnots
    ):
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfFootId].translation
        rhFootPos0 = self.rdata.oMf[self.rhFootId].translation
        lfFootPos0 = self.rdata.oMf[self.lfFootId].translation
        lhFootPos0 = self.rdata.oMf[self.lhFootId].translation
        df = jumpLength[2] - rfFootPos0[2]
        rfFootPos0[2] = 0.0
        rhFootPos0[2] = 0.0
        lfFootPos0[2] = 0.0
        lhFootPos0[2] = 0.0
        comRef = (rfFootPos0 + rhFootPos0 + lfFootPos0 + lhFootPos0) / 4
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2].item()

        loco3dModel = []
        takeOff = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
            )
            for k in range(groundKnots)
        ]
        flyingUpPhase = [
            self.createSwingFootModel(
                timeStep,
                [],
                np.array([jumpLength[0], jumpLength[1], jumpLength[2] + jumpHeight])
                * (k + 1)
                / flyingKnots
                + comRef,
            )
            for k in range(flyingKnots)
        ]
        flyingDownPhase = []
        for k in range(flyingKnots):
            flyingDownPhase += [self.createSwingFootModel(timeStep, [])]

        f0 = jumpLength
        footTask = [
            [self.lfFootId, pinocchio.SE3(np.eye(3), lfFootPos0 + f0)],
            [self.rfFootId, pinocchio.SE3(np.eye(3), rfFootPos0 + f0)],
            [self.lhFootId, pinocchio.SE3(np.eye(3), lhFootPos0 + f0)],
            [self.rhFootId, pinocchio.SE3(np.eye(3), rhFootPos0 + f0)],
        ]
        landingPhase = [
            self.createFootSwitchModel(
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
                footTask,
                False,
            )
        ]
        f0[2] = df
        landed = [
            self.createSwingFootModel(
                timeStep,
                [self.lfFootId, self.rfFootId, self.lhFootId, self.rhFootId],
                comTask=comRef + f0,
            )
            for k in range(groundKnots)
        ]
        loco3dModel += takeOff
        loco3dModel += flyingUpPhase
        loco3dModel += flyingDownPhase
        loco3dModel += landingPhase
        loco3dModel += landed

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createFootstepModels(
        self,
        comPos0,
        feetPos0,
        stepLength,
        stepHeight,
        timeStep,
        numKnots,
        supportFootIds,
        swingFootIds,
    ):
        """Action models for a footstep phase.

        :param comPos0, initial CoM position
        :param feetPos0: initial position of the swinging feet
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: time step
        :param numKnots: number of knots for the footstep phase
        :param supportFootIds: Ids of the supporting feet
        :param swingFootIds: Ids of the swinging foot
        :return footstep action models
        """
        numLegs = len(supportFootIds) + len(swingFootIds)
        comPercentage = float(len(swingFootIds)) / numLegs

        # Action models for the foot swing
        footSwingModel = []
        for k in range(numKnots):
            swingFootTask = []
            for i, p in zip(swingFootIds, feetPos0):
                # Defining a foot swing task given the step length
                # resKnot = numKnots % 2
                phKnots = numKnots / 2
                if k < phKnots:
                    dp = np.array(
                        [stepLength * (k + 1) / numKnots, 0.0, stepHeight * k / phKnots]
                    )
                elif k == phKnots:
                    dp = np.array([stepLength * (k + 1) / numKnots, 0.0, stepHeight])
                else:
                    dp = np.array(
                        [
                            stepLength * (k + 1) / numKnots,
                            0.0,
                            stepHeight * (1 - float(k - phKnots) / phKnots),
                        ]
                    )
                tref = p + dp

                swingFootTask += [[i, pinocchio.SE3(np.eye(3), tref)]]

            comTask = (
                np.array([stepLength * (k + 1) / numKnots, 0.0, 0.0]) * comPercentage
                + comPos0
            )
            footSwingModel += [
                self.createSwingFootModel(
                    timeStep,
                    supportFootIds,
                    comTask=comTask,
                    swingFootTask=swingFootTask,
                )
            ]

        # Action model for the foot switch
        footSwitchModel = self.createFootSwitchModel(supportFootIds, swingFootTask)

        # Updating the current foot position for next step
        comPos0 += [stepLength * comPercentage, 0.0, 0.0]
        for p in feetPos0:
            p += [stepLength, 0.0, 0.0]
        return [*footSwingModel, footSwitchModel]

    def createSwingFootModel(
        self, timeStep, supportFootIds, comTask=None, swingFootTask=None
    ):
        """Action model for a swing foot phase.

        :param timeStep: step duration of the action model
        :param supportFootIds: Ids of the constrained feet
        :param comTask: CoM task
        :param swingFootTask: swinging foot task
        :return action model for a swing foot phase
        """
        # Creating a 3D multi-contact model, and then including the supporting
        # foot
        if self._fwddyn:
            nu = self.actuation.nu
        else:
            nu = self.state.nv + 3 * len(supportFootIds)
        contactModel = crocoddyl.ContactModelMultiple(self.state, nu)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ContactModel3D(
                self.state,
                i,
                np.array([0.0, 0.0, 0.0]),
                pinocchio.LOCAL_WORLD_ALIGNED,
                nu,
                np.array([0.0, 50.0]),
            )
            contactModel.addContact(
                self.rmodel.frames[i].name + "_contact", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, nu)
        if isinstance(comTask, np.ndarray):
            comResidual = crocoddyl.ResidualModelCoMPosition(self.state, comTask, nu)
            comTrack = crocoddyl.CostModelResidual(self.state, comResidual)
            costModel.addCost("comTrack", comTrack, 1e6)
        for i in supportFootIds:
            cone = crocoddyl.FrictionCone(self.Rsurf, self.mu, 4, False)
            coneResidual = crocoddyl.ResidualModelContactFrictionCone(
                self.state, i, cone, nu, self._fwddyn
            )
            coneActivation = crocoddyl.ActivationModelQuadraticBarrier(
                crocoddyl.ActivationBounds(cone.lb, cone.ub)
            )
            frictionCone = crocoddyl.CostModelResidual(
                self.state, coneActivation, coneResidual
            )
            costModel.addCost(
                self.rmodel.frames[i].name + "_frictionCone", frictionCone, 1e1
            )
        if swingFootTask is not None:
            for i in swingFootTask:
                frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                    self.state, i[0], i[1].translation, nu
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, frameTranslationResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e6
                )

        stateWeights = np.array(
            [0.0] * 3
            + [500.0] * 3
            + [0.01] * (self.rmodel.nv - 6)
            + [10.0] * 6
            + [1.0] * (self.rmodel.nv - 6)
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, nu
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        if self._fwddyn:
            ctrlResidual = crocoddyl.ResidualModelControl(self.state, nu)
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        else:
            ctrlResidual = crocoddyl.ResidualModelJointEffort(
                self.state, self.actuation, nu
            )
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        costModel.addCost("stateReg", stateReg, 1e1)
        costModel.addCost("ctrlReg", ctrlReg, 1e-1)

        lb = np.concatenate(
            [self.state.lb[1 : self.state.nv + 1], self.state.lb[-self.state.nv :]]
        )
        ub = np.concatenate(
            [self.state.ub[1 : self.state.nv + 1], self.state.ub[-self.state.nv :]]
        )
        stateBoundsResidual = crocoddyl.ResidualModelState(self.state, nu)
        stateBoundsActivation = crocoddyl.ActivationModelQuadraticBarrier(
            crocoddyl.ActivationBounds(lb, ub)
        )
        stateBounds = crocoddyl.CostModelResidual(
            self.state, stateBoundsActivation, stateBoundsResidual
        )
        costModel.addCost("stateBounds", stateBounds, 1e3)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        if self._fwddyn:
            dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
                self.state, self.actuation, contactModel, costModel, 0.0, True
            )
        else:
            dmodel = crocoddyl.DifferentialActionModelContactInvDynamics(
                self.state, self.actuation, contactModel, costModel
            )
        if self._control == "one":
            control = crocoddyl.ControlParametrizationModelPolyOne(nu)
        elif self._control == "rk4":
            control = crocoddyl.ControlParametrizationModelPolyTwoRK(
                nu, crocoddyl.RKType.four
            )
        elif self._control == "rk3":
            control = crocoddyl.ControlParametrizationModelPolyTwoRK(
                nu, crocoddyl.RKType.three
            )
        else:
            control = crocoddyl.ControlParametrizationModelPolyZero(nu)
        if self._integrator == "euler":
            model = crocoddyl.IntegratedActionModelEuler(dmodel, control, timeStep)
        elif self._integrator == "rk4":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.four, timeStep
            )
        elif self._integrator == "rk3":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.three, timeStep
            )
        elif self._integrator == "rk2":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.two, timeStep
            )
        else:
            model = crocoddyl.IntegratedActionModelEuler(dmodel, control, timeStep)
        return model

    def createFootSwitchModel(self, supportFootIds, swingFootTask, pseudoImpulse=False):
        """Action model for a foot switch phase.

        :param supportFootIds: Ids of the constrained feet
        :param swingFootTask: swinging foot task
        :param pseudoImpulse: true for pseudo-impulse models, otherwise it uses the
            impulse model
        :return action model for a foot switch phase
        """
        if pseudoImpulse:
            return self.createPseudoImpulseModel(supportFootIds, swingFootTask)
        else:
            return self.createImpulseModel(supportFootIds, swingFootTask)

    def createPseudoImpulseModel(self, supportFootIds, swingFootTask):
        """Action model for pseudo-impulse models.

        A pseudo-impulse model consists of adding high-penalty cost for the contact
        velocities.
        :param supportFootIds: Ids of the constrained feet
        :param swingFootTask: swinging foot task
        :return pseudo-impulse differential action model
        """
        # Creating a 3D multi-contact model, and then including the supporting
        # foot
        if self._fwddyn:
            nu = self.actuation.nu
        else:
            nu = self.state.nv + 3 * len(supportFootIds)
        contactModel = crocoddyl.ContactModelMultiple(self.state, nu)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ContactModel3D(
                self.state,
                i,
                np.array([0.0, 0.0, 0.0]),
                pinocchio.LOCAL_WORLD_ALIGNED,
                nu,
                np.array([0.0, 50.0]),
            )
            contactModel.addContact(
                self.rmodel.frames[i].name + "_contact", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, nu)
        for i in supportFootIds:
            cone = crocoddyl.FrictionCone(self.Rsurf, self.mu, 4, False)
            coneResidual = crocoddyl.ResidualModelContactFrictionCone(
                self.state, i, cone, nu, self._fwddyn
            )
            coneActivation = crocoddyl.ActivationModelQuadraticBarrier(
                crocoddyl.ActivationBounds(cone.lb, cone.ub)
            )
            frictionCone = crocoddyl.CostModelResidual(
                self.state, coneActivation, coneResidual
            )
            costModel.addCost(
                self.rmodel.frames[i].name + "_frictionCone", frictionCone, 1e1
            )
        if swingFootTask is not None:
            for i in swingFootTask:
                frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                    self.state, i[0], i[1].translation, nu
                )
                frameVelocityResidual = crocoddyl.ResidualModelFrameVelocity(
                    self.state,
                    i[0],
                    pinocchio.Motion.Zero(),
                    pinocchio.LOCAL_WORLD_ALIGNED,
                    nu,
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, frameTranslationResidual
                )
                impulseFootVelCost = crocoddyl.CostModelResidual(
                    self.state, frameVelocityResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e7
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_impulseVel",
                    impulseFootVelCost,
                    1e6,
                )

        stateWeights = np.array(
            [0.0] * 3
            + [500.0] * 3
            + [0.01] * (self.rmodel.nv - 6)
            + [10.0] * self.rmodel.nv
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, nu
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        if self._fwddyn:
            ctrlResidual = crocoddyl.ResidualModelControl(self.state, nu)
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        else:
            ctrlResidual = crocoddyl.ResidualModelJointEffort(
                self.state, self.actuation, nu
            )
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        costModel.addCost("stateReg", stateReg, 1e1)
        costModel.addCost("ctrlReg", ctrlReg, 1e-3)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        if self._fwddyn:
            dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
                self.state, self.actuation, contactModel, costModel, 0.0, True
            )
        else:
            dmodel = crocoddyl.DifferentialActionModelContactInvDynamics(
                self.state, self.actuation, contactModel, costModel
            )
        if self._integrator == "euler":
            model = crocoddyl.IntegratedActionModelEuler(dmodel, 0.0)
        elif self._integrator == "rk4":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, crocoddyl.RKType.four, 0.0
            )
        elif self._integrator == "rk3":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, crocoddyl.RKType.three, 0.0
            )
        elif self._integrator == "rk2":
            model = crocoddyl.IntegratedActionModelRK(dmodel, crocoddyl.RKType.two, 0.0)
        else:
            model = crocoddyl.IntegratedActionModelEuler(dmodel, 0.0)
        return model

    def createImpulseModel(
        self, supportFootIds, swingFootTask, JMinvJt_damping=1e-12, r_coeff=0.0
    ):
        """Action model for impulse models.

        An impulse model consists of describing the impulse dynamics against a set of
        contacts.
        :param supportFootIds: Ids of the constrained feet
        :param swingFootTask: swinging foot task
        :return impulse action model
        """
        # Creating a 3D multi-contact model, and then including the supporting foot
        impulseModel = crocoddyl.ImpulseModelMultiple(self.state)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ImpulseModel3D(
                self.state, i, pinocchio.LOCAL_WORLD_ALIGNED
            )
            impulseModel.addImpulse(
                self.rmodel.frames[i].name + "_impulse", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, 0)
        if swingFootTask is not None:
            for i in swingFootTask:
                frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                    self.state, i[0], i[1].translation, 0
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, frameTranslationResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e7
                )

        stateWeights = np.array(
            [1.0] * 6 + [10.0] * (self.rmodel.nv - 6) + [10.0] * self.rmodel.nv
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, 0
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        costModel.addCost("stateReg", stateReg, 1e1)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        model = crocoddyl.ActionModelImpulseFwdDynamics(
            self.state, impulseModel, costModel
        )
        model.JMinvJt_damping = JMinvJt_damping
        model.r_coeff = r_coeff
        return model


def plotSolution(solver, bounds=True, figIndex=1, figTitle="", show=True):
    import matplotlib.pyplot as plt

    xs, us, cs = [], [], []
    if bounds:
        us_lb, us_ub = [], []
        xs_lb, xs_ub = [], []

    def updateTrajectories(solver):
        xs.extend(solver.xs[:-1])
        for m, d in zip(solver.problem.runningModels, solver.problem.runningDatas):
            if hasattr(m, "differential"):
                cs.append(d.differential.multibody.pinocchio.com[0])
                us.append(d.differential.multibody.joint.tau)
                if bounds and isinstance(
                    m.differential, crocoddyl.DifferentialActionModelContactFwdDynamics
                ):
                    us_lb.extend([m.u_lb])
                    us_ub.extend([m.u_ub])
            else:
                cs.append(d.multibody.pinocchio.com[0])
                us.append(np.zeros(nu))
                if bounds:
                    us_lb.append(np.nan * np.ones(nu))
                    us_ub.append(np.nan * np.ones(nu))
            if bounds:
                xs_lb.extend([m.state.lb])
                xs_ub.extend([m.state.ub])

    if isinstance(solver, list):
        for s in solver:
            rmodel = solver[0].problem.runningModels[0].state.pinocchio
            nq, nv, nu = (
                rmodel.nq,
                rmodel.nv,
                solver[0].problem.runningModels[0].differential.actuation.nu,
            )
            updateTrajectories(s)
    else:
        rmodel = solver.problem.runningModels[0].state.pinocchio
        nq, nv, nu = (
            rmodel.nq,
            rmodel.nv,
            solver.problem.runningModels[0].differential.actuation.nu,
        )
        updateTrajectories(solver)

    # Getting the state and control trajectories
    nx = nq + nv
    X = [0.0] * nx
    U = [0.0] * nu
    if bounds:
        U_LB = [0.0] * nu
        U_UB = [0.0] * nu
        X_LB = [0.0] * nx
        X_UB = [0.0] * nx
    for i in range(nx):
        X[i] = [x[i] for x in xs]
        if bounds:
            X_LB[i] = [x[i] for x in xs_lb]
            X_UB[i] = [x[i] for x in xs_ub]
    for i in range(nu):
        U[i] = [u[i] for u in us]
        if bounds:
            U_LB[i] = [u[i] for u in us_lb]
            U_UB[i] = [u[i] for u in us_ub]

    # Plotting the joint positions, velocities and torques
    plt.figure(figIndex)
    plt.suptitle(figTitle)
    legJointNames = ["HAA", "HFE", "KFE"]
    # LF foot
    plt.subplot(4, 3, 1)
    plt.title("joint position [rad]")
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(7, 10))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(7, 10))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(7, 10))]
    plt.ylabel("LF")
    plt.legend()
    plt.subplot(4, 3, 2)
    plt.title("joint velocity [rad/s]")
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 6, nq + 9))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 6, nq + 9))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 6, nq + 9))]
    plt.ylabel("LF")
    plt.legend()
    plt.subplot(4, 3, 3)
    plt.title("joint torque [Nm]")
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(0, 3))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(0, 3))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(0, 3))]
    plt.ylabel("LF")
    plt.legend()

    # LH foot
    plt.subplot(4, 3, 4)
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(10, 13))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(10, 13))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(10, 13))]
    plt.ylabel("LH")
    plt.legend()
    plt.subplot(4, 3, 5)
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 9, nq + 12))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 9, nq + 12))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 9, nq + 12))]
    plt.ylabel("LH")
    plt.legend()
    plt.subplot(4, 3, 6)
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(3, 6))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(3, 6))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(3, 6))]
    plt.ylabel("LH")
    plt.legend()

    # RF foot
    plt.subplot(4, 3, 7)
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(13, 16))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(13, 16))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(13, 16))]
    plt.ylabel("RF")
    plt.legend()
    plt.subplot(4, 3, 8)
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 12, nq + 15))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 12, nq + 15))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 12, nq + 15))]
    plt.ylabel("RF")
    plt.legend()
    plt.subplot(4, 3, 9)
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(6, 9))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(6, 9))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(6, 9))]
    plt.ylabel("RF")
    plt.legend()

    # RH foot
    plt.subplot(4, 3, 10)
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(16, 19))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(16, 19))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(16, 19))]
    plt.ylabel("RH")
    plt.xlabel("knots")
    plt.legend()
    plt.subplot(4, 3, 11)
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 15, nq + 18))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 15, nq + 18))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 15, nq + 18))]
    plt.ylabel("RH")
    plt.xlabel("knots")
    plt.legend()
    plt.subplot(4, 3, 12)
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(9, 12))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(9, 12))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(9, 12))]
    plt.ylabel("RH")
    plt.legend()
    plt.xlabel("knots")

    plt.figure(figIndex + 1)
    plt.suptitle(figTitle)
    Cx = [c[0] for c in cs]
    Cy = [c[1] for c in cs]
    plt.plot(Cx, Cy)
    plt.title("CoM position")
    plt.xlabel("x [m]")
    plt.ylabel("y [m]")
    plt.grid(True)
    if show:
        plt.show()



import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl
from crocoddyl.utils.quadruped import SimpleQuadrupedalGaitProblem

T = int(sys.argv[1]) if (len(sys.argv) > 1) else int(5e3)  # number of trials
MAXITER = 1
CALLBACKS = False
WALKING = "walk" in sys.argv
TROTTING = "trot" in sys.argv
PACING = "pace" in sys.argv
BOUNDING = "bound" in sys.argv
JUMPING = "jump" in sys.argv

GAIT = "walking"  # 104 nodes
if WALKING:
    print("running walking benchmark ...")
    GAIT = "walking"  # 104 nodes
if TROTTING:
    print("running trotting benchmark ...")
    GAIT = "trotting"  # 54 nodes
if PACING:
    print("running pacing benchmark ...")
    GAIT = "pacing"  # 54 nodes
if BOUNDING:
    print("running bounding benchmark ...")
    GAIT = "bounding"  # 54 nodes
if JUMPING:
    print("running jumping benchmark ...")
    GAIT = "jumping"  # 61 nodes


def createProblem(gait_phase):
    robot_model = example_robot_data.load("hyq").model
    lfFoot, rfFoot, lhFoot, rhFoot = "lf_foot", "rf_foot", "lh_foot", "rh_foot"
    gait = SimpleQuadrupedalGaitProblem(robot_model, lfFoot, rfFoot, lhFoot, rhFoot)
    q0 = robot_model.referenceConfigurations["standing"].copy()
    v0 = pinocchio.utils.zero(robot_model.nv)
    x0 = np.concatenate([q0, v0])

    type_of_gait = next(iter(gait_phase.keys()))
    value = gait_phase[type_of_gait]
    if type_of_gait == "walking":
        # Creating a walking problem
        problem = gait.createWalkingProblem(
            x0,
            value["stepLength"],
            value["stepHeight"],
            value["timeStep"],
            value["stepKnots"],
            value["supportKnots"],
        )
    elif type_of_gait == "trotting":
        # Creating a trotting problem
        problem = gait.createTrottingProblem(
            x0,
            value["stepLength"],
            value["stepHeight"],
            value["timeStep"],
            value["stepKnots"],
            value["supportKnots"],
        )
    elif type_of_gait == "pacing":
        # Creating a pacing problem
        problem = gait.createPacingProblem(
            x0,
            value["stepLength"],
            value["stepHeight"],
            value["timeStep"],
            value["stepKnots"],
            value["supportKnots"],
        )
    elif type_of_gait == "bounding":
        # Creating a bounding problem
        problem = gait.createBoundingProblem(
            x0,
            value["stepLength"],
            value["stepHeight"],
            value["timeStep"],
            value["stepKnots"],
            value["supportKnots"],
        )
    elif type_of_gait == "jumping":
        # Creating a jumping problem
        problem = gait.createJumpingProblem(
            x0,
            value["jumpHeight"],
            value["jumpLength"],
            value["timeStep"],
            value["groundKnots"],
            value["flyingKnots"],
        )

    xs = [robot_model.defaultState] * (len(problem.runningModels) + 1)
    us = problem.quasiStatic([robot_model.defaultState] * problem.T)
    return xs, us, problem


def runDDPSolveBenchmark(xs, us, problem):
    solver = crocoddyl.SolverFDDP(problem)
    if CALLBACKS:
        solver.setCallbacks([crocoddyl.CallbackVerbose()])
    duration = []
    for _ in range(T):
        c_start = time.time()
        solver.solve(xs, us, MAXITER, False, 0.1)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_dur = sum(duration) / len(duration)
    min_dur = min(duration)
    max_dur = max(duration)
    return avrg_dur, min_dur, max_dur


def runShootingProblemCalcBenchmark(xs, us, problem):
    duration = []
    for _ in range(T):
        c_start = time.time()
        problem.calc(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_dur = sum(duration) / len(duration)
    min_dur = min(duration)
    max_dur = max(duration)
    return avrg_dur, min_dur, max_dur


def runShootingProblemCalcDiffBenchmark(xs, us, problem):
    duration = []
    for i in range(T):
        c_start = time.time()
        problem.calcDiff(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_dur = sum(duration) / len(duration)
    min_dur = min(duration)
    max_dur = max(duration)
    return avrg_dur, min_dur, max_dur


# Setting up all tasks
if GAIT == "walking":
    GAITPHASE = {
        "walking": {
            "stepLength": 0.25,
            "stepHeight": 0.25,
            "timeStep": 1e-2,
            "stepKnots": 25,
            "supportKnots": 2,
        }
    }
elif GAIT == "trotting":
    GAITPHASE = {
        "trotting": {
            "stepLength": 0.15,
            "stepHeight": 0.2,
            "timeStep": 1e-2,
            "stepKnots": 25,
            "supportKnots": 2,
        }
    }
elif GAIT == "pacing":
    GAITPHASE = {
        "pacing": {
            "stepLength": 0.15,
            "stepHeight": 0.2,
            "timeStep": 1e-2,
            "stepKnots": 25,
            "supportKnots": 2,
        }
    }
elif GAIT == "bounding":
    GAITPHASE = {
        "bounding": {
            "stepLength": 0.007,
            "stepHeight": 0.05,
            "timeStep": 1e-2,
            "stepKnots": 25,
            "supportKnots": 12,
        }
    }
elif GAIT == "jumping":
    GAITPHASE = {
        "jumping": {
            "jumpHeight": 0.15,
            "jumpLength": [0.0, 0.3, 0.0],
            "timeStep": 1e-2,
            "groundKnots": 10,
            "flyingKnots": 20,
        }
    }

xs, us, problem = createProblem(GAITPHASE)
print("NQ:", problem.terminalModel.state.nq)
print("Number of nodes:", problem.T)
avrg_dur, min_dur, max_dur = runDDPSolveBenchmark(xs, us, problem)
print(f"  FDDP.solve [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcBenchmark(xs, us, problem)
print(f"  ShootingProblem.calc [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcDiffBenchmark(xs, us, problem)
print(f"  ShootingProblem.calcDiff [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
